# -*- coding: utf-8 -*-
"""Serial Session implementation using PyUSB.


:copyright: 2014-2020 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""

import errno
from typing import Any, List, Tuple, Type, Union

from pyvisa import attributes, constants
from pyvisa.constants import ResourceAttribute, StatusCode
from pyvisa.rname import USBInstr, USBRaw

from .common import logger
from .sessions import Session, UnknownAttribute

try:
    import usb

    from .protocols import usbraw, usbtmc, usbutil
except ImportError as e:
    msg = "Please install PyUSB to use this resource type.\n%s"
    Session.register_unavailable(constants.InterfaceType.usb, "INSTR", msg % e)
    Session.register_unavailable(constants.InterfaceType.usb, "RAW", msg % e)
    raise

try:
    _ = usb.core.find()
except Exception as e:
    msg = (
        "PyUSB does not seem to be properly installed.\n"
        "Please refer to PyUSB documentation and \n"
        "install a suitable backend like \n"
        "libusb 0.1, libusb 1.0, libusbx, \n"
        "libusb-win32 or OpenUSB.\n%s" % e
    )
    Session.register_unavailable(constants.InterfaceType.usb, "INSTR", msg)
    Session.register_unavailable(constants.InterfaceType.usb, "RAW", msg)
    raise


class USBTimeoutException(Exception):
    """Exception used internally to indicate USB timeout."""

    pass


class USBSession(Session):
    """Base class for drivers working with usb devices via usb port using pyUSB."""

    # Override parsed to take into account the fact that this class is only used
    # for a specific kind of resource
    parsed: Union[USBInstr, USBRaw]

    #: Class to use when instantiating the interface
    _intf_cls: Union[Type[usbraw.USBRawDevice], Type[usbtmc.USBTMC]]

    @staticmethod
    def list_resources() -> List[str]:
        """Return list of resources for this type of USB device."""
        raise NotImplementedError

    @classmethod
    def get_low_level_info(cls) -> str:
        try:
            ver = usb.__version__
        except AttributeError:
            ver = "N/A"

        try:
            # noinspection PyProtectedMember
            backend = usb.core.find()._ctx.backend.__class__.__module__.split(".")[-1]
        except Exception:
            backend = "N/A"

        return "via PyUSB (%s). Backend: %s" % (ver, backend)

    def after_parsing(self) -> None:
        self.interface = self._intf_cls(
            int(self.parsed.manufacturer_id, 0),
            int(self.parsed.model_code, 0),
            self.parsed.serial_number,
        )

        self.attrs.update(
            {
                ResourceAttribute.manufacturer_id: int(self.parsed.manufacturer_id, 0),
                ResourceAttribute.model_code: int(self.parsed.model_code, 0),
                ResourceAttribute.usb_serial_number: self.parsed.serial_number,
                ResourceAttribute.usb_interface_number: int(
                    self.parsed.usb_interface_number
                ),
            }
        )

        for name, attr in (
            ("SEND_END_EN", ResourceAttribute.send_end_enabled),
            ("SUPPRESS_END_EN", ResourceAttribute.suppress_end_enabled),
            ("TERMCHAR", ResourceAttribute.termchar),
            ("TERMCHAR_EN", ResourceAttribute.termchar_enabled),
        ):
            attribute = getattr(constants, "VI_ATTR_" + name)
            self.attrs[attr] = attributes.AttributesByID[attribute].default

        # Force setting the timeout to get the proper value
        self.set_attribute(
            ResourceAttribute.timeout_value,
            attributes.AttributesByID[attribute].default,
        )

    def _get_timeout(self, attribute: ResourceAttribute) -> Tuple[int, StatusCode]:
        if self.interface:
            if self.interface.timeout == 2**32 - 1:
                self.timeout = None
            else:
                self.timeout = self.interface.timeout / 1000
        return super(USBSession, self)._get_timeout(attribute)

    def _set_timeout(self, attribute: ResourceAttribute, value: int) -> StatusCode:
        status = super(USBSession, self)._set_timeout(attribute, value)
        timeout = int(self.timeout * 1000) if self.timeout else 2**32 - 1
        timeout = min(timeout, 2**32 - 1)
        if self.interface:
            self.interface.timeout = timeout
        return status

    def read(self, count: int) -> Tuple[bytes, StatusCode]:
        """Reads data from device or interface synchronously.

        Corresponds to viRead function of the VISA library.

        Parameters
        -----------
        count : int
            Number of bytes to be read.

        Returns
        -------
        bytes
            Data read from the device
        StatusCode
            Return value of the library call.

        """

        def _usb_reader():
            """Data reader identifying usb timeout exception."""
            try:
                return self.interface.read(count)
            except usb.USBError as exc:
                if exc.errno in (errno.ETIMEDOUT, -errno.ETIMEDOUT):
                    raise USBTimeoutException()
                raise

        supress_end_en, _ = self.get_attribute(ResourceAttribute.suppress_end_enabled)

        if supress_end_en:
            raise ValueError(
                "VI_ATTR_SUPPRESS_END_EN == True is currently unsupported by pyvisa-py"
            )

        term_char, _ = self.get_attribute(ResourceAttribute.termchar)
        term_char_en, _ = self.get_attribute(ResourceAttribute.termchar_enabled)

        return self._read(
            _usb_reader,
            count,
            lambda current: True,  # USB always returns a complete message
            supress_end_en,
            term_char,
            term_char_en,
            USBTimeoutException,
        )

    def write(self, data: bytes) -> Tuple[int, StatusCode]:
        """Writes data to device or interface synchronously.

        Corresponds to viWrite function of the VISA library.

        Parameters
        ----------
        data : bytes
            Data to be written.

        Returns
        -------
        int
            Number of bytes actually transferred
        StatusCode
            Return value of the library call.

        """
        send_end, _ = self.get_attribute(ResourceAttribute.send_end_enabled)

        count = self.interface.write(data)

        return count, StatusCode.success

    def close(self):
        self.interface.close()
        return StatusCode.success

    def _get_attribute(
        self, attribute: constants.ResourceAttribute
    ) -> Tuple[Any, StatusCode]:
        """Get the value for a given VISA attribute for this session.

        Use to implement custom logic for attributes.

        Parameters
        ----------
        attribute : ResourceAttribute
            Attribute for which the state query is made

        Returns
        -------
        Any
            State of the queried attribute for a specified resource
        StatusCode
            Return value of the library call.

        """
        raise UnknownAttribute(attribute)

    def _set_attribute(
        self, attribute: constants.ResourceAttribute, attribute_state: Any
    ) -> StatusCode:
        """Sets the state of an attribute.

        Corresponds to viSetAttribute function of the VISA library.

        Parameters
        ----------
        attribute : constants.ResourceAttribute
            Attribute for which the state is to be modified. (Attributes.*)
        attribute_state : Any
            The state of the attribute to be set for the specified object.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        raise UnknownAttribute(attribute)


@Session.register(constants.InterfaceType.usb, "INSTR")
class USBInstrSession(USBSession):
    """Class for USBTMC devices."""

    # Override parsed to take into account the fact that this class is only used
    # for a specific kind of resource
    parsed: USBInstr

    #: Class to use when instantiating the interface
    _intf_cls = usbtmc.USBTMC

    @staticmethod
    def list_resources() -> List[str]:
        out = []
        fmt = (
            "USB%(board)s::%(manufacturer_id)s::%(model_code)s::"
            "%(serial_number)s::%(usb_interface_number)s::INSTR"
        )
        for dev in usbtmc.find_tmc_devices():
            intfc = usbutil.find_interfaces(
                dev, bInterfaceClass=0xFE, bInterfaceSubClass=3
            )
            try:
                intfc = intfc[0].index
            except (IndexError, AttributeError):
                intfc = 0

            try:
                serial = dev.serial_number
            except (NotImplementedError, ValueError):
                msg = (
                    "Found a device whose serial number cannot be read."
                    " The partial VISA resource name is: " + fmt
                )
                logger.warning(
                    msg,
                    {
                        "board": 0,
                        "manufacturer_id": dev.idVendor,
                        "model_code": dev.idProduct,
                        "serial_number": "???",
                        "usb_interface_number": intfc,
                    },
                )
                continue

            out.append(
                fmt
                % {
                    "board": 0,
                    "manufacturer_id": dev.idVendor,
                    "model_code": dev.idProduct,
                    "serial_number": serial,
                    "usb_interface_number": intfc,
                }
            )
        return out


@Session.register(constants.InterfaceType.usb, "RAW")
class USBRawSession(USBSession):
    """Class for RAW devices."""

    # Override parsed to take into account the fact that this class is only used
    # for a specific kind of resource
    parsed: USBRaw

    #: Class to use when instantiating the interface
    _intf_cls = usbraw.USBRawDevice

    @staticmethod
    def list_resources() -> List[str]:
        out = []
        fmt = (
            "USB%(board)s::%(manufacturer_id)s::%(model_code)s::"
            "%(serial_number)s::%(usb_interface_number)s::RAW"
        )
        for dev in usbraw.find_raw_devices():
            intfc = usbutil.find_interfaces(dev, bInterfaceClass=0xFF)
            try:
                intfc = intfc[0].index
            except (IndexError, AttributeError):
                intfc = 0

            try:
                serial = dev.serial_number
            except (NotImplementedError, ValueError, usb.USBError):
                msg = (
                    "Found a device whose serial number cannot be read."
                    " The partial VISA resource name is: " + fmt
                )
                logger.warning(
                    msg,
                    {
                        "board": 0,
                        "manufacturer_id": dev.idVendor,
                        "model_code": dev.idProduct,
                        "serial_number": "???",
                        "usb_interface_number": intfc,
                    },
                )
                continue

            out.append(
                fmt
                % {
                    "board": 0,
                    "manufacturer_id": dev.idVendor,
                    "model_code": dev.idProduct,
                    "serial_number": serial,
                    "usb_interface_number": intfc,
                }
            )
        return out



# -*- coding: utf-8 -*-
"""Pure Python backend for PyVISA.


:copyright: 2014-2020 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""

import sys

if sys.version_info >= (3, 8):
    from importlib.metadata import PackageNotFoundError, version
else:
    from importlib_metadata import PackageNotFoundError, version  # type: ignore

__version__ = "unknown"
try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    pass

# noqa: we need to import so that __init_subclass__() is executed once
from . import attributes  # noqa: F401
from .highlevel import PyVisaLibrary

WRAPPER_CLASS = PyVisaLibrary



# -*- coding: utf-8 -*-
"""Test the Resource manager."""

import pytest

from pyvisa.rname import ResourceName
from pyvisa.testsuite.keysight_assisted_tests import (
    RESOURCE_ADDRESSES,
    copy_func,
    require_virtual_instr,
)
from pyvisa.testsuite.keysight_assisted_tests.test_resource_manager import (
    TestResourceManager as BaseTestResourceManager,
    TestResourceParsing as BaseTestResourceParsing,
)


@require_virtual_instr
class TestPyResourceManager(BaseTestResourceManager):
    """ """

    def test_list_resource(self):
        """Test listing the available resources.
        The bot supports only TCPIP and of those resources we expect to be able
        to list only INSTR resources not SOCKET.
        """
        # Default settings
        resources = self.rm.list_resources()
        for v in (v for v in RESOURCE_ADDRESSES.values() if v.endswith("INSTR")):
            assert str(ResourceName.from_string(v)) in resources

    test_last_status = pytest.mark.xfail(
        copy_func(BaseTestResourceManager.test_last_status)
    )

    test_opening_resource_with_lock = pytest.mark.xfail(
        copy_func(BaseTestResourceManager.test_opening_resource_with_lock)
    )


@require_virtual_instr
class TestPyResourceParsing(BaseTestResourceParsing):
    """ """

    pass



# -*- coding: utf-8 -*-
"""Common code.

:copyright: 2014-2020 by PyVISA-sim Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""

import logging
from typing import Iterator, Optional

from pyvisa import logger

logger = logging.LoggerAdapter(logger, {"backend": "py"})  # type: ignore


class NamedObject(object):
    """A class to construct named sentinels."""

    #: Name used to identify the sentinel
    name: str

    def __init__(self, name) -> None:
        self.name = name

    def __repr__(self) -> str:
        return "<%s>" % self.name

    __str__ = __repr__


def int_to_byte(val):
    return val.to_bytes(1, "big")


# TODO(anyone): This is copypasta from `pyvisa-sim` project - find a way to
#   reduce duplication, probably in that project instead of here.
def _create_bitmask(bits: int) -> int:
    """Create a bitmask for the given number of bits."""
    mask = (1 << bits) - 1
    return mask


# TODO(anyone): This is copypasta from `pyvisa-sim` project - find a way to
#   reduce duplication, probably in that project instead of here.
def iter_bytes(
    data: bytes, data_bits: Optional[int] = None, send_end: Optional[bool] = None
) -> Iterator[bytes]:
    """Clip values to the correct number of bits per byte.
    Serial communication may use from 5 to 8 bits.
    Parameters
    ----------
    data : The data to clip as a byte string.
    data_bits : How many bits per byte should be sent. Clip to this many bits.
        For example: data_bits=5: 0xff (0b1111_1111) --> 0x1f (0b0001_1111).
        Acceptable range is 5 to 8, inclusive. Values above 8 will be clipped to 8.
        This maps to the VISA attribute VI_ATTR_ASRL_DATA_BITS.
    send_end :
        If None (the default), apply the mask that is determined by data_bits.
        If False, apply the mask and set the highest (post-mask) bit to 0 for
        all bytes.
        If True, apply the mask and set the highest (post-mask) bit to 0 for
        all bytes except for the final byte, which has the highest bit set to 1.
    References
    ----------
    + https://www.ivifoundation.org/downloads/Architecture%20Specifications/vpp43_2022-05-19.pdf,
    + https://www.ni.com/docs/en-US/bundle/ni-visa/page/ni-visa/vi_attr_asrl_data_bits.html,
    + https://www.ni.com/docs/en-US/bundle/ni-visa/page/ni-visa/vi_attr_asrl_end_out.html

    """
    if send_end and data_bits is None:
        raise ValueError("'send_end' requires a valid 'data_bits' value.")

    if data_bits is None:
        for d in data:
            yield bytes([d])
    else:
        if data_bits <= 0:
            raise ValueError("'data_bits' cannot be zero or negative")
        if data_bits > 8:
            data_bits = 8

        if send_end is None:
            # only apply the mask
            mask = _create_bitmask(data_bits)
            for d in data:
                yield bytes([d & mask])
        elif bool(send_end) is False:
            # apply the mask and set highest bits to 0
            # This is effectively the same has reducing the mask by 1 bit.
            mask = _create_bitmask(data_bits - 1)
            for d in data:
                yield bytes([d & mask])
        elif bool(send_end) is True:
            # apply the mask and set highest bits to 0
            # This is effectively the same has reducing the mask by 1 bit.
            mask = _create_bitmask(data_bits - 1)
            for d in data[:-1]:
                yield bytes([d & mask])
            # except for the last byte which has it's highest bit set to 1.
            last_byte = data[-1]
            highest_bit = 1 << (data_bits - 1)
            yield bytes([(last_byte & mask) | highest_bit])
        else:
            raise ValueError(f"Unknown 'send_end' value '{send_end}'")



# -*- coding: utf-8 -*-
"""Serial Session implementation using PySerial.


:copyright: 2014-2020 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""

import sys
from typing import Any, List, Tuple

from pyvisa import attributes, constants, logger, rname
from pyvisa.constants import (
    BufferOperation,
    ResourceAttribute,
    SerialTermination,
    StatusCode,
)

from . import common
from .sessions import Session, UnknownAttribute

try:
    import serial
    from serial.tools.list_ports import comports
except ImportError as e:
    Session.register_unavailable(
        constants.InterfaceType.asrl,
        "INSTR",
        "Please install PySerial (>=3.0) to use this resource type.\n%s" % e,
    )
    raise

IS_WIN = sys.platform == "win32"


def to_state(boolean_input: bool) -> constants.LineState:
    """Convert a boolean input into a LineState value."""
    if boolean_input:
        return constants.LineState.asserted
    return constants.LineState.unasserted


@Session.register(constants.InterfaceType.asrl, "INSTR")
class SerialSession(Session):
    """A serial Session that uses PySerial to do the low level communication."""

    # Override parsed to take into account the fact that this class is only used
    # for a specific kind of resource
    parsed: rname.ASRLInstr

    @staticmethod
    def list_resources() -> List[str]:
        return [
            "ASRL%s::INSTR" % (port[0][3:] if IS_WIN else port[0])
            for port in comports()
        ]

    @classmethod
    def get_low_level_info(cls) -> str:
        try:
            ver = serial.VERSION
        except AttributeError:
            ver = "N/A"

        return "via PySerial (%s)" % ver

    def after_parsing(self) -> None:
        self.interface = serial.serial_for_url(
            ("COM" if IS_WIN else "") + self.parsed.board,
            timeout=self.timeout,
            write_timeout=self.timeout,
        )

        for name in (
            "ASRL_END_IN",
            "ASRL_END_OUT",
            "SEND_END_EN",
            "TERMCHAR",
            "TERMCHAR_EN",
            "SUPPRESS_END_EN",
        ):
            attribute = getattr(constants, "VI_ATTR_" + name)
            self.attrs[attribute] = attributes.AttributesByID[attribute].default

    def _get_timeout(self, attribute: ResourceAttribute) -> Tuple[int, StatusCode]:
        if self.interface:
            self.timeout = self.interface.timeout
        return super(SerialSession, self)._get_timeout(attribute)

    def _set_timeout(self, attribute: ResourceAttribute, value: int) -> StatusCode:
        status = super(SerialSession, self)._set_timeout(attribute, value)
        if self.interface:
            self.interface.timeout = self.timeout
            self.interface.write_timeout = self.timeout
        return status

    def close(self) -> StatusCode:
        self.interface.close()
        return StatusCode.success

    def read(self, count: int) -> Tuple[bytes, StatusCode]:
        """Reads data from device or interface synchronously.

        Corresponds to viRead function of the VISA library.

        Parameters
        -----------
        count : int
            Number of bytes to be read.

        Returns
        -------
        bytes
            Data read from the device
        StatusCode
            Return value of the library call.

        """
        end_in, _ = self.get_attribute(ResourceAttribute.asrl_end_in)
        suppress_end_en, _ = self.get_attribute(ResourceAttribute.suppress_end_enabled)

        reader = lambda: self.interface.read(1)  # noqa: E731

        if end_in == SerialTermination.none:
            checker = lambda current: False  # noqa: E731

        elif end_in == SerialTermination.last_bit:
            mask = 2**self.interface.bytesize
            checker = lambda current: bool(current[-1] & mask)  # noqa: E731

        elif end_in == SerialTermination.termination_char:
            end_char, _ = self.get_attribute(ResourceAttribute.termchar)

            checker = lambda current: current[-1] == end_char  # noqa: E731

        else:
            raise ValueError("Unknown value for VI_ATTR_ASRL_END_IN: %s" % end_in)

        return self._read(
            reader,
            count,
            checker,
            suppress_end_en,
            None,
            False,
            serial.SerialTimeoutException,
        )

    def write(self, data: bytes) -> Tuple[int, StatusCode]:
        """Writes data to device or interface synchronously.

        Corresponds to viWrite function of the VISA library.

        Parameters
        ----------
        data : bytes
            Data to be written.

        Returns
        -------
        int
            Number of bytes actually transferred
        StatusCode
            Return value of the library call.

        """
        logger.debug("Serial.write %r" % data)
        send_end, _ = self.get_attribute(ResourceAttribute.send_end_enabled)
        end_out, _ = self.get_attribute(ResourceAttribute.asrl_end_out)
        data_bits, _ = self.get_attribute(constants.ResourceAttribute.asrl_data_bits)

        if end_out == SerialTermination.none:
            pass
        elif end_out == SerialTermination.last_bit:
            data = b"".join(common.iter_bytes(data, data_bits, send_end))
        elif end_out == SerialTermination.termination_char:
            term_char, _ = self.get_attribute(ResourceAttribute.termchar)
            data = b"".join(common.iter_bytes(data, data_bits, send_end=None))
            data = data + common.int_to_byte(term_char)
        elif end_out == SerialTermination.termination_break:
            data = b"".join(common.iter_bytes(data, data_bits, send_end=None))
        else:
            raise ValueError("Unknown value for VI_ATTR_ASRL_END_OUT: %s" % end_out)

        try:
            count = self.interface.write(data)

            if end_out == SerialTermination.termination_break:
                logger.debug("Serial.sendBreak")
                self.interface.sendBreak()

            return count, StatusCode.success

        except serial.SerialTimeoutException:
            return 0, StatusCode.error_timeout

    def flush(self, mask: BufferOperation) -> StatusCode:
        """Flush the specified buffers.

        The buffers can be associated with formatted I/O operations and/or
        serial communication.

        Corresponds to viFlush function of the VISA library.

        Parameters
        ----------
        mask : constants.BufferOperation
            Specifies the action to be taken with flushing the buffer.
            The values can be combined using the | operator. However multiple
            operations on a single buffer cannot be combined.

        Returns
        -------
        constants.StatusCode
            Return value of the library call.

        """
        if mask & BufferOperation.discard_read_buffer:
            self.interface.reset_input_buffer()
        if (
            mask & BufferOperation.flush_write_buffer
            or mask & BufferOperation.flush_transmit_buffer
        ):
            self.interface.flush()
        if (
            mask & BufferOperation.discard_write_buffer
            or mask & BufferOperation.discard_transmit_buffer
        ):
            self.interface.reset_output_buffer()

        return StatusCode.success

    def _get_attribute(  # noqa: C901
        self, attribute: constants.ResourceAttribute
    ) -> Tuple[Any, StatusCode]:
        """Get the value for a given VISA attribute for this session.

        Use to implement custom logic for attributes.

        Parameters
        ----------
        attribute : ResourceAttribute
            Attribute for which the state query is made

        Returns
        -------
        Any
            State of the queried attribute for a specified resource
        StatusCode
            Return value of the library call.

        """
        if attribute == constants.VI_ATTR_ASRL_ALLOW_TRANSMIT:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_AVAIL_NUM:
            return self.interface.inWaiting(), StatusCode.success

        elif attribute == constants.VI_ATTR_ASRL_BAUD:
            return self.interface.baudrate, StatusCode.success

        elif attribute == constants.VI_ATTR_ASRL_BREAK_LEN:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_BREAK_STATE:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_CONNECTED:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_CTS_STATE:
            return to_state(self.interface.getCTS()), StatusCode.success

        elif attribute == constants.VI_ATTR_ASRL_DATA_BITS:
            return self.interface.bytesize, StatusCode.success

        elif attribute == constants.VI_ATTR_ASRL_DCD_STATE:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_DISCARD_NULL:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_DSR_STATE:
            return to_state(self.interface.getDSR()), StatusCode.success

        elif attribute == constants.VI_ATTR_ASRL_DTR_STATE:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_FLOW_CNTRL:
            return (
                (
                    self.interface.xonxoff * constants.VI_ASRL_FLOW_XON_XOFF
                    | self.interface.rtscts * constants.VI_ASRL_FLOW_RTS_CTS
                    | self.interface.dsrdtr * constants.VI_ASRL_FLOW_DTR_DSR
                ),
                StatusCode.success,
            )

        elif attribute == constants.VI_ATTR_ASRL_PARITY:
            parity = self.interface.parity
            if parity == serial.PARITY_NONE:
                return constants.Parity.none, StatusCode.success
            elif parity == serial.PARITY_EVEN:
                return constants.Parity.even, StatusCode.success
            elif parity == serial.PARITY_ODD:
                return constants.Parity.odd, StatusCode.success
            elif parity == serial.PARITY_MARK:
                return constants.Parity.mark, StatusCode.success
            elif parity == serial.PARITY_SPACE:
                return constants.Parity.space, StatusCode.success

            raise Exception("Unknown parity value: %r" % parity)

        elif attribute == constants.VI_ATTR_ASRL_RI_STATE:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_RTS_STATE:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_STOP_BITS:
            bits = self.interface.stopbits
            if bits == serial.STOPBITS_ONE:
                return constants.StopBits.one, StatusCode.success
            elif bits == serial.STOPBITS_ONE_POINT_FIVE:
                return constants.StopBits.one_and_a_half, StatusCode.success
            elif bits == serial.STOPBITS_TWO:
                return constants.StopBits.two, StatusCode.success

            raise Exception("Unknown bits value: %r" % bits)

        elif attribute == constants.VI_ATTR_ASRL_XOFF_CHAR:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_INTF_TYPE:
            return constants.InterfaceType.asrl, StatusCode.success

        raise UnknownAttribute(attribute)

    def _set_attribute(  # noqa: C901
        self, attribute: constants.ResourceAttribute, attribute_state: Any
    ) -> StatusCode:
        """Sets the state of an attribute.

        Corresponds to viSetAttribute function of the VISA library.

        Parameters
        ----------
        attribute : constants.ResourceAttribute
            Attribute for which the state is to be modified. (Attributes.*)
        attribute_state : Any
            The state of the attribute to be set for the specified object.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        if attribute == constants.VI_ATTR_ASRL_ALLOW_TRANSMIT:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_BAUD:
            self.interface.baudrate = attribute_state
            return StatusCode.success

        elif attribute == constants.VI_ATTR_ASRL_BREAK_LEN:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_BREAK_STATE:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_CONNECTED:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_DATA_BITS:
            self.interface.bytesize = attribute_state
            return StatusCode.success

        elif attribute == constants.VI_ATTR_ASRL_DCD_STATE:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_DISCARD_NULL:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_DSR_STATE:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_DTR_STATE:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_FLOW_CNTRL:
            if not isinstance(attribute_state, int):
                return StatusCode.error_nonsupported_attribute_state

            if not 0 <= attribute_state < 8:
                return StatusCode.error_nonsupported_attribute_state

            try:
                self.interface.xonxoff = (
                    attribute_state & constants.VI_ASRL_FLOW_XON_XOFF
                )
                self.interface.rtscts = attribute_state & constants.VI_ASRL_FLOW_RTS_CTS
                self.interface.dsrdtr = attribute_state & constants.VI_ASRL_FLOW_DTR_DSR
                return StatusCode.success
            except Exception:
                return StatusCode.error_nonsupported_attribute_state

        elif attribute == constants.VI_ATTR_ASRL_PARITY:
            if attribute_state == constants.Parity.none:
                self.interface.parity = serial.PARITY_NONE
                return StatusCode.success

            elif attribute_state == constants.Parity.even:
                self.interface.parity = serial.PARITY_EVEN
                return StatusCode.success

            elif attribute_state == constants.Parity.odd:
                self.interface.parity = serial.PARITY_ODD
                return StatusCode.success

            elif attribute_state == serial.PARITY_MARK:
                self.interface.parity = serial.PARITY_MARK
                return StatusCode.success

            elif attribute_state == constants.Parity.space:
                self.interface.parity = serial.PARITY_SPACE
                return StatusCode.success

            return StatusCode.error_nonsupported_attribute_state

        elif attribute == constants.VI_ATTR_ASRL_RI_STATE:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_RTS_STATE:
            raise NotImplementedError

        elif attribute == constants.VI_ATTR_ASRL_STOP_BITS:
            if attribute_state == constants.StopBits.one:
                self.interface.stopbits = serial.STOPBITS_ONE
                return StatusCode.success

            if attribute_state == constants.StopBits.one_and_a_half:
                self.interface.stopbits = serial.STOPBITS_ONE_POINT_FIVE
                return StatusCode.success

            if attribute_state == constants.StopBits.two:
                self.interface.stopbits = serial.STOPBITS_TWO
                return StatusCode.success

            return StatusCode.error_nonsupported_attribute_state

        elif attribute == constants.VI_ATTR_ASRL_XOFF_CHAR:
            raise NotImplementedError

        raise UnknownAttribute(attribute)



"""Test loading resources.


:copyright: 2014-2023 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""

import ctypes

from pyvisa.constants import InterfaceType
from pyvisa.testsuite import BaseTestCase
from pyvisa_py.sessions import Session


class TestSessions(BaseTestCase):
    """Test generic property of PyVisaLibrary."""

    def test_sessions(self):
        available = [d for d, _ in Session.iter_valid_session_classes()]
        missing = [d for d, _ in Session.iter_session_classes_issues()]

        expected = [(InterfaceType.tcpip, "INSTR"), (InterfaceType.tcpip, "SOCKET")]
        exp_missing = []
        usbs = [(InterfaceType.usb, "INSTR"), (InterfaceType.usb, "RAW")]
        try:
            import usb

            _ = usb.core.find()

            expected.extend(usbs)
        except Exception:
            exp_missing.extend(usbs)

        gpibs = [(InterfaceType.gpib, "INSTR"), (InterfaceType.gpib, "INTFC")]
        try:
            try:
                from gpib_ctypes import gpib
                from gpib_ctypes.Gpib import Gpib
                from gpib_ctypes.gpib.gpib import _lib as gpib_lib
            except ImportError:
                import gpib  # noqa
                from Gpib import Gpib  # noqa
            else:
                # Add some extra binding not available by default
                extra_funcs = [
                    ("ibcac", [ctypes.c_int, ctypes.c_int], ctypes.c_int),
                    ("ibgts", [ctypes.c_int, ctypes.c_int], ctypes.c_int),
                    ("ibpct", [ctypes.c_int], ctypes.c_int),
                ]
                for name, argtypes, restype in extra_funcs:
                    libfunction = gpib_lib[name]
                    libfunction.argtypes = argtypes
                    libfunction.restype = restype

            expected.extend(gpibs)
        except Exception:
            exp_missing.extend(gpibs)

        asrl = (InterfaceType.asrl, "INSTR")
        try:
            import serial  # noqa

            expected.append(asrl)
        except Exception:
            exp_missing.append(asrl)

        vicp = (InterfaceType.vicp, "INSTR")
        try:
            import pyvicp  # noqa

            expected.append(vicp)
        except Exception:
            exp_missing.append(vicp)

        assert sorted(available) == sorted(expected)
        assert sorted(missing) == sorted(exp_missing)



# -*- coding: utf-8 -*-
import os
import unittest

# Set the environment variable to use PyVISA-py as backend
os.environ["PYVISA_LIBRARY"] = "@py"


def testsuite():
    """A testsuite that has all the pyvisa-py tests."""
    return unittest.TestLoader().discover(os.path.dirname(__file__))


def main():
    """Runs the testsuite as command line application."""
    try:
        unittest.main()
    except Exception as e:
        print("Error: %s" % e)


def run() -> unittest.TestResult:
    """Run all tests."""
    test_runner = unittest.TextTestRunner()
    return test_runner.run(testsuite())



# -*- coding: utf-8 -*-
"""Additional Attributes for specific use with the pyvisa-py package.

For additional information and VISA attributes see pyvisa.constants

:copyright: 2014-2020 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.
"""

from pyvisa import constants
from pyvisa.attributes import AttrVI_ATTR_TCPIP_KEEPALIVE as former_keepalive


class AttrVI_ATTR_TCPIP_KEEPALIVE(former_keepalive):
    """Requests that a TCP/IP provider enable the use of keep-alive packets.

    Altering the standard PyVISA attribute to also work on INSTR sessions as
    they are using sockets in pyvisa-py as well.

    After the system detects that a connection was dropped, VISA returns a lost
    connection error code on subsequent I/O calls on the session. The time required
    for the system to detect that the connection was dropped is dependent on the
    system and is not settable.

    """

    resources = [
        (constants.InterfaceType.tcpip, "SOCKET"),
        (constants.InterfaceType.tcpip, "INSTR"),
    ]



# -*- coding: utf-8 -*-
"""Serial Session implementation using PyUSB.

See the following link for more information about USB.

http://www.beyondlogic.org/usbnutshell/usb5.shtml

This file is an offspring of the Lantz Project.

:copyright: 2014-2020 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""

from fnmatch import fnmatch

import usb
from usb.util import find_descriptor as usb_find_desc

ClassCodes = {
    0x00: ("Device", "Use class information in the Interface Descriptors"),
    0x01: ("Interface", "Audio"),
    0x02: ("Both", "Communications and CDC Control"),
    0x03: ("Interface", "HID (Human Interface Device)"),
    0x05: ("Interface", "Physical"),
    0x06: ("Interface", "Image"),
    0x07: ("Interface", "Printer"),
    0x08: ("Interface", "Mass Storage"),
    0x09: ("Device", "Hub"),
    0x0A: ("Interface", "CDC-Data"),
    0x0B: ("Interface", "Smart Card"),
    0x0D: ("Interface", "Content Security"),
    0x0E: ("Interface", "Video"),
    0x0F: ("Interface", "Personal Healthcare"),
    0x10: ("Interface", "Audio/Video Devices"),
    0xDC: ("Both", "Diagnostic Device"),
    0xE0: ("Interface", "Wireless Controller"),
    0xEF: ("Both", "Miscellaneous"),
    0xFE: ("Interface", "Application Specific"),
    0xFF: ("Both", "Vendor Specific"),
}

# None is 0xxx
AllCodes = {
    (0x00, 0x00, 0x00): "Use class code info from Interface Descriptors",
    (0x01, None, None): "Audio device",
    (0x02, None, None): "Communication device class",
    (0x03, None, None): "HID device class",
    (0x05, None, None): "Physical device class",
    (0x06, 0x01, 0x01): "Still Imaging device",
    (0x07, None, None): "Printer device",
    (0x08, None, None): "Mass Storage device",
    (0x09, 0x00, 0x00): "Full speed Hub",
    (0x09, 0x00, 0x01): "Hi-speed hub with single TT",
    (0x09, 0x00, 0x02): "Hi-speed hub with multiple TTs",
    (0x0A, None, None): "CDC data device",
    (0x0B, None, None): "Smart Card device",
    (0x0D, 0x00, 0x00): "Content Security device",
    (0x0E, None, None): "Video device",
    (0x0F, None, None): "Personal Healthcare device",
    (0x10, 0x01, 0x00): "Control Interface",
    (0x10, 0x02, 0x00): "Data Video Streaming Interface",
    (0x10, 0x03, 0x00): "VData Audio Streaming Interface",
    (0xDC, 0x01, 0x01): "USB2 Compliance Device",
    (0xE0, 0x01, 0x01): "Bluetooth Programming Interface.",
    (0xE0, 0x01, 0x02): "UWB Radio Control Interface.",
    (0xE0, 0x01, 0x03): "Remote NDIS",
    (0xE0, 0x01, 0x04): "Bluetooth AMP Controller.",
    (0xE0, 0x2, 0x01): "Host Wire Adapter Control/Data interface.",
    (0xE0, 0x2, 0x02): "Device Wire Adapter Control/Data interface.",
    (0xE0, 0x2, 0x03): "Device Wire Adapter Isochronous interface.",
    (0xEF, 0x01, 0x01): "Active Sync device.",
    (0xEF, 0x01, 0x02): "Palm Sync. This class code can be used in either "
    "Device or Interface Descriptors.",
    (0xEF, 0x02, 0x01): "Interface Association Descriptor.",
    (0xEF, 0x02, 0x02): "Wire Adapter Multifunction Peripheral programming interface.",
    (0xEF, 0x03, 0x01): "Cable Based Association Framework.",
    (0xEF, 0x04, 0x01): "RNDIS over Ethernet. Connecting a host to the Internet via "
    "Ethernet mobile device. The device appears to the host as an"
    "Ethernet gateway device. This class code may only be used in "
    "Interface Descriptors.",
    (0xEF, 0x04, 0x02): "RNDIS over WiFi. Connecting a host to the Internet via WiFi "
    "enabled mobile device.  The device represents itself to the host"
    "as an 802.11 compliant network device. This class code may only"
    "be used in Interface Descriptors.",
    (0xEF, 0x04, 0x03): "RNDIS over WiMAX. Connecting a host to the Internet via WiMAX "
    "enabled mobile device.  The device is represented to the host "
    "as an 802.16 network device. This class code may only be used "
    "in Interface Descriptors.",
    (
        0xEF,
        0x04,
        0x04,
    ): "RNDIS over WWAN. Connecting a host to the Internet via a device "
    "using mobile broadband, i.e. WWAN (GSM/CDMA). This class code may "
    "only be used in Interface Descriptors.",
    (
        0xEF,
        0x04,
        0x05,
    ): "RNDIS for Raw IPv4. Connecting a host to the Internet using raw "
    "IPv4 via non-Ethernet mobile device.  Devices that provide raw "
    "IPv4, not in an Ethernet packet, may use this form to in lieu of "
    "other stock types. "
    "This class code may only be used in Interface Descriptors.",
    (
        0xEF,
        0x04,
        0x06,
    ): "RNDIS for Raw IPv6. Connecting a host to the Internet using raw "
    "IPv6 via non-Ethernet mobile device.  Devices that provide raw "
    "IPv6, not in an Ethernet packet, may use this form to in lieu of "
    "other stock types. "
    "This class code may only be used in Interface Descriptors.",
    (
        0xEF,
        0x04,
        0x07,
    ): "RNDIS for GPRS. Connecting a host to the Internet over GPRS mobile "
    "device using the device,Äôs cellular radio.",
    (0xEF, 0x05, 0x00): "USB3 Vision Control Interface",
    (0xEF, 0x05, 0x01): "USB3 Vision Event Interface",
    (0xEF, 0x05, 0x02): "USB3 Vision Streaming Interface",
    (0xFE, 0x01, 0x01): "Device Firmware Upgrade.",
    (0xFE, 0x02, 0x00): "IRDA Bridge device.",
    (0xFE, 0x03, 0x00): "USB Test and Measurement Device.",
    (
        0xFE,
        0x03,
        0x01,
    ): "USB Test and Measurement Device conforming to the USBTMC USB488 Subclass",
    (0xFF, None, None): "Vendor specific",
}


def ep_attributes(ep):
    c = ep.bmAttributes
    attrs = []
    tp = c & usb.ENDPOINT_TYPE_MASK
    if tp == usb.ENDPOINT_TYPE_CONTROL:
        attrs.append("Control")
    elif tp == usb.ENDPOINT_TYPE_ISOCHRONOUS:
        attrs.append("Isochronous")
    elif tp == usb.ENDPOINT_TYPE_BULK:
        attrs.append("Bulk")
    elif tp == usb.ENDPOINT_TYPE_INTERRUPT:
        attrs.append("Interrupt")

    sync = (c & 12) >> 2
    if sync == 0:
        attrs.append("No sync")
    elif sync == 1:
        attrs.append("Async")
    elif sync == 2:
        attrs.append("Adaptive")
    elif sync == 3:
        attrs.append("Sync")
    usage = (c & 48) >> 4
    if usage == 0:
        attrs.append("Data endpoint")
    elif usage == 1:
        attrs.append("Feedback endpoint")
    elif usage == 2:
        attrs.append("Subordinate Feedback endpoint")
    elif usage == 3:
        attrs.append("Reserved")

    return ", ".join(attrs)


def find_devices(
    vendor=None, product=None, serial_number=None, custom_match=None, **kwargs
):
    """Find connected USB devices matching certain keywords.

    Wildcards can be used for vendor, product and serial_number.

    :param vendor: name or id of the vendor (manufacturer)
    :param product: name or id of the product
    :param serial_number: serial number.
    :param custom_match: callable returning True or False that takes a device as only input.
    :param kwargs: other properties to match. See usb.core.find
    :return:
    """
    kwargs = kwargs or {}
    attrs = {}
    if isinstance(vendor, str):
        attrs["manufacturer"] = vendor
    elif vendor is not None:
        kwargs["idVendor"] = vendor

    if isinstance(product, str):
        attrs["product"] = product
    elif product is not None:
        kwargs["idProduct"] = product

    if serial_number:
        attrs["serial_number"] = str(serial_number)

    if attrs:

        def cm(dev):
            if custom_match is not None and not custom_match(dev):
                return False
            for attr, pattern in attrs.items():
                try:
                    value = getattr(dev, attr)
                except (NotImplementedError, ValueError):
                    return False
                if not fnmatch(value.lower(), pattern.lower()):
                    return False
            return True

    else:
        cm = custom_match

    return usb.core.find(find_all=True, custom_match=cm, **kwargs)


def find_interfaces(device, **kwargs):
    """
    :param device:
    :return:
    """
    interfaces = []
    try:
        for cfg in device:
            try:
                interfaces.extend(usb_find_desc(cfg, find_all=True, **kwargs))
            except Exception:
                pass
    except Exception:
        pass
    return interfaces


def find_endpoint(interface, direction, type):
    ep = usb_find_desc(
        interface,
        custom_match=lambda e: usb.util.endpoint_direction(e.bEndpointAddress)
        == direction
        and usb.util.endpoint_type(e.bmAttributes) == type,
    )
    return ep


def _patch_endpoint(ep, log_func=print):
    _read = ep.read
    _write = ep.write

    def new_read(*args, **kwargs):
        log_func("---")
        log_func("reading from {}".format(ep.bEndpointAddress))
        log_func("args: {}".format(args))
        log_func("kwargs: {}".format(kwargs))
        ret = _read(*args, **kwargs)
        log_func("returned", ret)
        log_func("---")
        return ret

    def new_write(*args, **kwargs):
        log_func("---")
        log_func("writing to {}".format(ep.bEndpointAddress))
        log_func("args: {}".format(args))
        log_func("kwargs: {}".format(kwargs))
        ret = _write(*args, **kwargs)
        log_func("returned", ret)
        log_func("---")
        return ret

    ep.read = new_read
    ep.write = new_write



# -*- coding: utf-8 -*-
"""Base Session class.


:copyright: 2014-2020 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""

import abc
import time
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
)

from pyvisa import attributes, constants, logger, rname
from pyvisa.constants import ResourceAttribute, StatusCode
from pyvisa.typing import VISARMSession

from .common import int_to_byte

#: Type var used when typing register.
T = TypeVar("T", bound=Type["Session"])


class UnknownAttribute(Exception):
    """Custom exception signaling a VISA attribute is not supported."""

    def __init__(self, attribute: constants.ResourceAttribute) -> None:
        self.attribute = attribute

    def __str__(self) -> str:
        attr = self.attribute
        if isinstance(attr, int):
            try:
                name = attributes.AttributesByID[attr].visa_name
            except KeyError:
                name = "Name not found"

            return "Unknown attribute %s (%s - %s)" % (attr, hex(attr), name)

        return "Unknown attribute %s" % attr

    __repr__ = __str__


class Session(metaclass=abc.ABCMeta):
    """A base class for Session objects.

    Just makes sure that common methods are defined and information is stored.

    Parameters
    ----------
    resource_manager_session : VISARMSession
        Session handle of the parent Resource Manager
    resource_name : str
        Name of the resource this session is communicating with
    parsed : rname.ResourceName, optional
        Parsed representation of the resource name. The default is False meaning
        that the provided resource name will be parsed.

    """

    @abc.abstractmethod
    def _get_attribute(
        self, attribute: constants.ResourceAttribute
    ) -> Tuple[Any, StatusCode]:
        """Get the value for a given VISA attribute for this session.

        Use to implement custom logic for attributes.

        Parameters
        ----------
        attribute : constants.ResourceAttribute
            Resource attribute for which the state query is made

        Returns
        -------
        Any
            State of the queried attribute for a specified resource
        constants.StatusCode
            Return value of the library call.

        """

    @abc.abstractmethod
    def _set_attribute(
        self, attribute: constants.ResourceAttribute, attribute_state: Any
    ) -> StatusCode:
        """Set the attribute_state value for a given VISA attribute for this session.

        Use to implement custom logic for attributes.

        Parameters
        ----------
        attribute : constants.ResourceAttribute
            Resource attribute for which the state query is made.
        attribute_state : Any
            Value to which to set the attribute.

        Returns
        -------
        StatusCode
            The return value of the library call.

        """

    @abc.abstractmethod
    def close(self) -> StatusCode:
        """Close the session.

        Use it to do final clean ups.

        """

    #: Session handle of the parent Resource Manager
    resource_manager_session: VISARMSession

    #: Name of the resource this session is communicating with
    resource_name: str

    #: Parsed representation of the resource name.
    parsed: rname.ResourceName

    #: Session type as (Interface Type, Resource Class)
    session_type: Tuple[constants.InterfaceType, str]

    #: Timeout in milliseconds to use when opening the resource.
    open_timeout: Optional[int]

    #: Value of the timeout in seconds used for general operation
    timeout: Optional[float]

    #: Used as a place holder for the object doing the lowlevel communication.
    interface: Any

    #: Used for attributes not handled by the underlying interface.
    #: Values are get or set automatically by get_attribute and set_attribute
    attrs: Dict[constants.ResourceAttribute, Any]

    #: Maps (Interface Type, Resource Class) to Python class encapsulating that
    #: resource.
    #: dict[(Interface Type, Resource Class) , Session]
    _session_classes: ClassVar[
        Dict[Tuple[constants.InterfaceType, str], Type["Session"]]
    ] = {}

    @staticmethod
    def list_resources() -> List[str]:
        """List the resources available for the resource class."""
        return []

    @classmethod
    def get_low_level_info(cls) -> str:
        """Get info about the backend used by the session."""
        return ""

    @classmethod
    def iter_valid_session_classes(
        cls,
    ) -> Iterator[Tuple[Tuple[constants.InterfaceType, str], Type["Session"]]]:
        """Iterator over valid sessions classes infos."""
        for key, val in cls._session_classes.items():
            if not issubclass(val, UnavailableSession):
                yield key, val

    @classmethod
    def iter_session_classes_issues(
        cls,
    ) -> Iterator[Tuple[Tuple[constants.InterfaceType, str], str]]:
        """Iterator over invalid sessions classes (i.e. those with import errors)."""
        for key, val in cls._session_classes.items():
            if issubclass(val, UnavailableSession):
                yield key, getattr(val, "session_issue")

    @classmethod
    def get_session_class(
        cls, interface_type: constants.InterfaceType, resource_class: str
    ) -> Type["Session"]:
        """Get the session class for a given interface type and resource class.

        Parameters
        ----------
        interface_type : constants.InterfaceType
            Type of interface.
        resource_class : str
            Class of resource.

        Returns
        -------
        Sessions
            Session subclass the most appropriate for the resource.

        """
        try:
            return cls._session_classes[(interface_type, resource_class)]
        except KeyError:
            raise ValueError(
                "No class registered for %s, %s" % (interface_type, resource_class)
            )

    @classmethod
    def register(
        cls, interface_type: constants.InterfaceType, resource_class: str
    ) -> Callable[[T], T]:
        """Register a session class for a given interface type and resource class.

        Parameters
        ----------
        interface_type : constants.InterfaceType
            Type of interface.
        resource_class : str
            Class of the resource

        Returns
        -------
        Callable[[T], T]
            Decorator function to register a session subclass.

        """

        def _internal(python_class):
            if (interface_type, resource_class) in cls._session_classes:
                logger.warning(
                    "%s is already registered in the "
                    "ResourceManager. Overwriting with %s",
                    (interface_type, resource_class),
                    python_class,
                )

            python_class.session_type = (interface_type, resource_class)
            cls._session_classes[(interface_type, resource_class)] = python_class
            return python_class

        return _internal

    @classmethod
    def register_unavailable(
        cls, interface_type: constants.InterfaceType, resource_class: str, msg: str
    ) -> None:
        """Register that no session class exists.

        This creates a fake session that will raise a ValueError if called.

        Parameters
        ----------
        interface_type : constants.InterfaceType
            Type of interface.
        resource_class : str
            Class of the resource
        msg : str
            Message detailing why no session class exists for this particular
            interface type, resource class pair.

        Returns
        -------
        Type[Session]
            Fake session.

        """

        class _internal(UnavailableSession):
            #: Message detailing why no session is available.
            session_issue = msg

        if (interface_type, resource_class) in cls._session_classes:
            logger.warning(
                "%s is already registered in the ResourceManager. "
                "Overwriting with unavailable %s",
                (interface_type, resource_class),
                msg,
            )

        cls._session_classes[(interface_type, resource_class)] = _internal

    def __init__(
        self,
        resource_manager_session: VISARMSession,
        resource_name: str,
        parsed: Optional[rname.ResourceName] = None,
        open_timeout: Optional[int] = None,
    ) -> None:
        if parsed is None:
            parsed = rname.parse_resource_name(resource_name)

        self.parsed = parsed
        self.open_timeout = open_timeout

        #: Used as a place holder for the object doing the lowlevel communication.
        self.interface = None

        #: Used for attributes not handled by the underlying interface.
        #: Values are get or set automatically by get_attribute and
        #: set_attribute
        #: Add your own by overriding after_parsing.
        self.attrs = {
            ResourceAttribute.resource_manager_session: resource_manager_session,
            ResourceAttribute.resource_name: str(parsed),
            ResourceAttribute.resource_class: parsed.resource_class,
            ResourceAttribute.interface_type: parsed.interface_type_const,
            ResourceAttribute.timeout_value: (self._get_timeout, self._set_timeout),
        }

        #: Timeout expressed in second or None for the absence of a timeout.
        #: The default value is set when calling self.set_attribute(attr, default_timeout)
        self.timeout = None

        #: Set the default timeout from constants
        attr = ResourceAttribute.timeout_value
        default_timeout = attributes.AttributesByID[attr].default
        self.set_attribute(attr, default_timeout)

        self.after_parsing()

    def after_parsing(self) -> None:
        """Override this method to provide custom initialization code, to be
        called after the resource name is properly parsed

        ResourceSession can register resource specific attributes handling of
        them into self.attrs.
        It is also possible to change handling of already registered common
        attributes. List of attributes is available in pyvisa package:
        * name is in constants module as: VI_ATTR_<NAME>
        * validity of attribute for resource is defined module attributes,
        AttrVI_ATTR_<NAME>.resources

        For static (read only) values, simple readonly and also readwrite
        attributes simplified construction can be used:
        `    self.attrs[constants.VI_ATTR_<NAME>] = 100`
        or
        `    self.attrs[constants.VI_ATTR_<NAME>] = <self.variable_name>`

        For more complex handling of attributes, it is possible to register
        getter and/or setter. When Null is used, NotSupported error is
        returned.
        Getter has same signature as see Session._get_attribute and setter has
        same signature as see Session._set_attribute. (It is possible to
        register also see Session._get_attribute and see Session._set_attribute
        as getter/setter). Getter and Setter are registered as tuple.
        For readwrite attribute:
        `    self.attrs[constants.VI_ATTR_<NAME>] = (<getter_name>,
                                                     <setter_name>)`
        For readonly attribute:
        `    self.attrs[constants.VI_ATTR_<NAME>] = (<getter_name>, None)`
        For reusing of see Session._get_attribute and see
        Session._set_attribute
        `    self.attrs[constants.VI_ATTR_<NAME>] = (self._get_attribute,
                                                     self._set_attribute)`

        """
        pass

    def write(self, data: bytes) -> Tuple[int, StatusCode]:
        """Writes data to device or interface synchronously.

        Corresponds to viWrite function of the VISA library.

        Parameters
        ----------
        data : bytes
            Data to be written.

        Returns
        -------
        int
            Number of bytes actually transferred
        StatusCode
            Return value of the library call.

        """
        raise NotImplementedError

    def read(self, count: int) -> Tuple[bytes, StatusCode]:
        """Reads data from device or interface synchronously.

        Corresponds to viRead function of the VISA library.

        Parameters
        -----------
        count : int
            Number of bytes to be read.

        Returns
        -------
        bytes
            Data read from the device
        StatusCode
            Return value of the library call.

        """
        raise NotImplementedError()

    def clear(self) -> StatusCode:
        """Clears a device.

        Corresponds to viClear function of the VISA library.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        return StatusCode.error_nonsupported_operation

    def flush(self, mask: constants.BufferOperation) -> StatusCode:
        """Flush the specified buffers.

        The buffers can be associated with formatted I/O operations and/or
        serial communication.

        Corresponds to viFlush function of the VISA library.

        Parameters
        ----------
        mask : constants.BufferOperation
            Specifies the action to be taken with flushing the buffer.
            The values can be combined using the | operator. However multiple
            operations on a single buffer cannot be combined.

        Returns
        -------
        constants.StatusCode
            Return value of the library call.

        """
        raise NotImplementedError

    def read_stb(self) -> Tuple[int, StatusCode]:
        """Reads a status byte of the service request.

        Corresponds to viReadSTB function of the VISA library.

        Returns
        -------
        int
            Service request status byte
        StatusCode
            Return value of the library call.

        """
        return 0, StatusCode.error_nonsupported_operation

    def lock(
        self,
        lock_type: constants.Lock,
        timeout: int,
        requested_key: Optional[str] = None,
    ):
        """Establishes an access mode to the specified resources.

        Corresponds to viLock function of the VISA library.

        Parameters
        ----------
        lock_type : constants.Lock
            Specifies the type of lock requested.
        timeout : int
            Absolute time period (in milliseconds) that a resource waits to get
            unlocked by the locking session before returning an error.
        requested_key : Optional[str], optional
            Requested locking key in the case of a shared lock. For an exclusive
            lock it should be None.

        Returns
        -------
        Optional[str]
            Key that can then be passed to other sessions to share the lock, or
            None for an exclusive lock.
        constants.StatusCode
            Return value of the library call.

        """
        return "", StatusCode.error_nonsupported_operation

    def unlock(self) -> StatusCode:
        """Relinquishes a lock for the specified resource.

        Corresponds to viUnlock function of the VISA library.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        return StatusCode.error_nonsupported_operation

    def gpib_command(self, command_byte: bytes) -> Tuple[int, StatusCode]:
        """Write GPIB command bytes on the bus.

        Corresponds to viGpibCommand function of the VISA library.

        Parameters
        ----------
        data : bytes
            Data to write.

        Returns
        -------
        int
            Number of written bytes
        constants.StatusCode
            Return value of the library call.

        """
        return 0, StatusCode.error_nonsupported_operation

    def assert_trigger(self, protocol: constants.TriggerProtocol) -> StatusCode:
        """Assert software or hardware trigger.

        Corresponds to viAssertTrigger function of the VISA library.

        Parameters
        ----------
        protocol : constants.TriggerProtocol
            Trigger protocol to use during assertion.

        Returns
        -------
        constants.StatusCode
            Return value of the library call.

        """
        raise NotImplementedError

    def gpib_send_ifc(self) -> StatusCode:
        """Pulse the interface clear line (IFC) for at least 100 microseconds.

        Corresponds to viGpibSendIFC function of the VISA library.

        Returns
        -------
        constants.StatusCode
            Return value of the library call.

        """
        return StatusCode.error_nonsupported_operation

    def gpib_control_ren(
        self, mode: constants.RENLineOperation
    ) -> constants.StatusCode:
        """Controls the state of the GPIB Remote Enable (REN) interface line.

        Optionally the remote/local state of the device can also be set.

        Corresponds to viGpibControlREN function of the VISA library.

        Parameters
        ----------
        mode : constants.RENLineOperation
            State of the REN line and optionally the device remote/local state.

        Returns
        -------
        constants.StatusCode
            Return value of the library call.

        """
        return StatusCode.error_nonsupported_operation

    def gpib_control_atn(self, mode: constants.ATNLineOperation) -> StatusCode:
        """Specifies the state of the ATN line and the local active controller state.

        Corresponds to viGpibControlATN function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        mode : constants.ATNLineOperation
            State of the ATN line and optionally the local active controller state.

        Returns
        -------
        constants.StatusCode
            Return value of the library call.

        """
        return StatusCode.error_nonsupported_operation

    def gpib_pass_control(
        self, primary_address: int, secondary_address: int
    ) -> StatusCode:
        """Tell a GPIB device to become controller in charge (CIC).

        Corresponds to viGpibPassControl function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        primary_address : int
            Primary address of the GPIB device to which you want to pass control.
        secondary_address : int
            Secondary address of the targeted GPIB device.
            If the targeted device does not have a secondary address, this parameter
            should contain the value Constants.VI_NO_SEC_ADDR.

        Returns
        -------
        constants.StatusCode
            Return value of the library call.

        """
        return StatusCode.error_nonsupported_operation

    def get_attribute(self, attribute: ResourceAttribute) -> Tuple[Any, StatusCode]:
        """Get the value for a given VISA attribute for this session.

        Does a few checks before and calls before dispatching to
        `_get_attribute`.

        Parameters
        ----------
        attribute : ResourceAttribute
            Resource attribute for which the state query is made

        Returns
        -------
        Any
            The state of the queried attribute for a specified resource.
        StatusCode
            Return value of the library call.

        """

        # Check if the attribute value is defined.
        try:
            attr = attributes.AttributesByID[attribute]
        except KeyError:
            return 0, StatusCode.error_nonsupported_attribute

        # Check if the attribute is defined for this session type.
        if not attr.in_resource(self.session_type):
            return 0, StatusCode.error_nonsupported_attribute

        # Check if reading the attribute is allowed.
        if not attr.read:
            raise Exception("Do not now how to handle write only attributes.")

        # First try to answer those attributes that are registered in
        # self.attrs, see Session.after_parsing
        if attribute in self.attrs:
            value = self.attrs[attribute]
            status = StatusCode.success
            if isinstance(value, tuple):
                getter = value[0]
                value, status = (
                    getter(attribute)
                    if getter
                    else (0, StatusCode.error_nonsupported_attribute)
                )
            return value, status

        # Dispatch to `_get_attribute`, which must be implemented by subclasses

        try:
            return self._get_attribute(attribute)
        except UnknownAttribute as e:
            logger.exception(str(e))
            return 0, StatusCode.error_nonsupported_attribute

    def set_attribute(
        self, attribute: ResourceAttribute, attribute_state: Any
    ) -> StatusCode:
        """Set the attribute_state value for a given VISA attribute for this
        session.

        Does a few checks before and calls before dispatching to
        `_set_attribute`.

        Parameters
        ----------
        attribute : ResourceAttribute
            Resource attribute for which the state query is made.
        attribute_state : Any
            Value.

        Returns
        -------
        StatusCode
            The return value of the library call.

        """
        # Check if the attribute value is defined.
        try:
            attr = attributes.AttributesByID[attribute]
        except KeyError:
            return StatusCode.error_nonsupported_attribute

        # Check if the attribute is defined for this session type.
        if not attr.in_resource(self.session_type):
            return StatusCode.error_nonsupported_attribute

        # Check if writing the attribute is allowed.
        if not attr.write:
            return StatusCode.error_attribute_read_only

        # First try to answer those attributes that are registered in
        # self.attrs, see Session.after_parsing
        if attribute in self.attrs:
            value = self.attrs[attribute]
            status = StatusCode.success
            if isinstance(value, tuple):
                setter = value[1]
                status = (
                    setter(attribute, attribute_state)
                    if setter
                    else StatusCode.error_nonsupported_attribute
                )
            else:
                self.attrs[attribute] = attribute_state
            return status

        # Dispatch to `_set_attribute`, which must be implemented by subclasses

        try:
            return self._set_attribute(attribute, attribute_state)
        except ValueError:
            return StatusCode.error_nonsupported_attribute_state
        except NotImplementedError:
            e = UnknownAttribute(attribute)
            logger.exception(str(e))
            return StatusCode.error_nonsupported_attribute
        except UnknownAttribute as e:
            logger.exception(str(e))
            return StatusCode.error_nonsupported_attribute

    def _read(
        self,
        reader: Callable[[], bytes],
        count: int,
        end_indicator_checker: Callable[[bytes], bool],
        suppress_end_en: bool,
        termination_char: Optional[int],
        termination_char_en: bool,
        timeout_exception: Type[Exception],
    ) -> Tuple[bytes, StatusCode]:
        """Reads data from device or interface synchronously.

        Corresponds to viRead function of the VISA library.

        Parameters
        ----------
        reader : Callable[[], bytes]
            Function to read one or more bytes.
        count : int
            Number of bytes to be read.
        end_indicator_checker : Callable[[bytes], bool]
            Function to check if the message is complete.
        suppress_end_en : bool
            Suppress end.
        termination_char : int
            Stop reading if this character is received.
        termination_char_en : bool
            Is termination char enabled.
        timeout_exception : Type[Exception]
            Exception to capture time out for the given interface.

        Returns
        -------
        bytes
            Data read from the resource.
        StatusCode
            Return value of the library call.

        """
        # NOTE: Some interfaces return not only a single byte but a complete
        # block for each read therefore we must handle the case that the
        # termination character is in the middle of the  block or that the
        # maximum number of bytes is exceeded

        # Turn the termination_char store as an int in VISA attribute in a byte
        term_char = (
            int_to_byte(termination_char) if termination_char is not None else b""
        )

        finish_time = None if self.timeout is None else (time.time() + self.timeout)
        out = bytearray()
        while True:
            try:
                current = reader()
            except timeout_exception:
                return out, StatusCode.error_timeout

            if current:
                out.extend(current)
                end_indicator_received = end_indicator_checker(current)
                if end_indicator_received:
                    if not suppress_end_en:
                        # RULE 6.1.1
                        return bytes(out), StatusCode.success
                else:
                    if termination_char_en and (term_char in current):
                        # RULE 6.1.2
                        # Return everything up to and including the termination
                        # character
                        return (
                            bytes(out[: out.index(term_char) + 1]),
                            StatusCode.success_termination_character_read,
                        )
                    elif len(out) >= count:
                        # RULE 6.1.3
                        # Return at most the number of bytes requested
                        return (bytes(out[:count]), StatusCode.success_max_count_read)

            if finish_time and time.time() > finish_time:
                return bytes(out), StatusCode.error_timeout

    def _get_timeout(self, attribute: ResourceAttribute) -> Tuple[int, StatusCode]:
        """Returns timeout calculated value from python way to VI_ way

        In VISA, the timeout is expressed in milliseconds or using the
        constants VI_TMO_INFINITE or VI_TMO_IMMEDIATE.

        In Python we store it as either None (VI_TMO_INFINITE), 0
        (VI_TMO_IMMEDIATE) or as a floating point number in seconds.

        """
        if self.timeout is None:
            ret_value = constants.VI_TMO_INFINITE
        elif self.timeout == 0:
            ret_value = constants.VI_TMO_IMMEDIATE
        else:
            ret_value = int(self.timeout * 1000.0)
        return ret_value, StatusCode.success

    def _set_timeout(self, attribute: ResourceAttribute, value: int):
        """Sets timeout calculated value from python way to VI_ way

        In VISA, the timeout is expressed in milliseconds or using the
        constants VI_TMO_INFINITE or VI_TMO_IMMEDIATE.

        In Python we store it as either None (VI_TMO_INFINITE), 0
        (VI_TMO_IMMEDIATE) or as a floating point number in seconds.

        """
        if value == constants.VI_TMO_INFINITE:
            self.timeout = None
        elif value == constants.VI_TMO_IMMEDIATE:
            self.timeout = 0
        else:
            self.timeout = value / 1000.0
        return StatusCode.success


class UnavailableSession(Session):
    session_issue: ClassVar[str]

    def __init__(self, *args, **kwargs) -> None:
        raise ValueError(self.session_issue)

    def _get_attribute(self, attr):
        raise NotImplementedError()

    def _set_attribute(self, attr, value):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()



"""Test creating a resource manager using PyVISA-Py as a backend.


:copyright: 2014-2023 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""

import pytest

from pyvisa import ResourceManager
from pyvisa.testsuite import BaseTestCase


class TestSerial(BaseTestCase):
    """Test generic property of PyVisaLibrary."""

    serial = pytest.importorskip("serial", reason="PySerial not installed")

    def test_serial(self):
        """Test loop://"""
        msg = b"Test01234567890"

        available = ["loop://"]
        expected = []
        exp_missing = []
        missing = {}

        rm = ResourceManager("@py")
        try:
            dut = rm.open_resource("ASRLloop://::INSTR")
            print("opened")
            dut.timeout = 3000
            dut.read_termination = "\r\n"
            dut.write_termination = "\r\n"
            dut.write(str(msg))
            ret_val = dut.read()
            if str(msg) == ret_val:
                expected = ["loop://"]

        except Exception:
            exp_missing = ["loop://"]

        assert sorted(available) == sorted(expected)
        assert sorted(missing) == sorted(exp_missing)



"""Test creating a resource manager using PyVISA-Py as a backend.


:copyright: 2014-2023 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""

from pyvisa.highlevel import list_backends
from pyvisa.testsuite import BaseTestCase
from pyvisa_py import highlevel


class TestPyVisaLibrary(BaseTestCase):
    """Test generic property of PyVisaLibrary."""

    def test_list_backends(self):
        """Test listing backends."""
        assert "py" in list_backends()

    def test_debug_info(self):
        """Test generating debug infos for PyVISA-py."""
        infos = highlevel.PyVisaLibrary.get_debug_info()
        for key in ("Version",):
            assert key in infos



# -*- coding: utf-8 -*-
"""Test relying on the Keysight virtual instrument.

The PyVISA builbot is connected to a fake instrument implemented using the
Keysight Virtual Instrument IO Test software.

For this part of the testsuite to be run, you need to set the
PYVISA_KEYSIGHT_VIRTUAL_INSTR environment value.

See pyvisa/testsuite/keysight_assisted_tests/__init__.py for more details.

:copyright: 2014-2020 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""



# -*- coding: utf-8 -*-
"""TCPIP Session implementation using Python Standard library.


:copyright: 2014-2020 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""

import ipaddress
import random
import select
import socket
import time
import warnings
from typing import Any, Dict, List, Optional, Tuple, Type

from pyvisa import attributes, constants, errors, rname
from pyvisa.constants import BufferOperation, ResourceAttribute, StatusCode

from . import common
from .protocols import hislip, rpc, vxi11
from .sessions import Session, UnknownAttribute, VISARMSession

# Let psutil be optional dependency
try:
    import psutil  # type: ignore
except ImportError:
    psutil = None

# Let zeroconf be optional dependency
try:
    import zeroconf  # type: ignore
except ImportError:
    zeroconf = None  # type: ignore

# Let pyvicp be an optional dependency
try:
    import pyvicp  # type: ignore
except ImportError:
    pyvicp = None  # type: ignore


# Conversion between VXI11 error codes and VISA status
# TODO this is so far a best guess, in particular 6 and 29 are likely wrong
VXI11_ERRORS_TO_VISA = {
    0: StatusCode.success,  # no_error
    1: StatusCode.error_invalid_format,  # syntax_error
    3: StatusCode.error_connection_lost,  # device_no_accessible
    4: StatusCode.error_invalid_access_key,  # invalid_link_identifier
    5: StatusCode.error_invalid_parameter,  # parameter_error
    6: StatusCode.error_handler_not_installed,  # channel_not_established
    8: StatusCode.error_nonsupported_operation,  # operation_not_supported
    9: StatusCode.error_allocation,  # out_of_resources
    11: StatusCode.error_resource_locked,  # device_locked_by_another_link
    12: StatusCode.error_session_not_locked,  # no_lock_held_by_this_link
    15: StatusCode.error_timeout,  # io_timeout
    17: StatusCode.error_io,  # io_error
    23: StatusCode.error_abort,  # abort
    29: StatusCode.error_window_already_mapped,  # channel_already_established
}


@Session.register(constants.InterfaceType.tcpip, "INSTR")
class TCPIPInstrSession(Session):
    """A class to dispatch to VXI11 or HiSLIP, based on the protocol."""

    def __new__(
        cls,
        resource_manager_session: VISARMSession,
        resource_name: str,
        parsed=None,
        open_timeout: Optional[int] = None,
    ):
        newcls: Type

        if parsed is None:
            parsed = rname.parse_resource_name(resource_name)

        if parsed.lan_device_name.lower().startswith("hislip"):
            newcls = TCPIPInstrHiSLIP

        else:
            newcls = TCPIPInstrVxi11

        return newcls(resource_manager_session, resource_name, parsed, open_timeout)

    @staticmethod
    def list_resources(wait_time=1.0) -> List[str]:
        return TCPIPInstrVxi11.list_resources() + TCPIPInstrHiSLIP.list_resources()

    @classmethod
    def get_low_level_info(cls) -> str:
        vxi11 = "ok" if psutil is not None else "partial (psutil not installed)"
        hislip = "ok" if zeroconf is not None else "disabled (zeroconf not installed)"
        return (
            "\n         Resource discovery:"
            f"\n         - VXI-11: {vxi11}"
            f"\n         - hislip: {hislip}"
        )


class TCPIPInstrHiSLIP(Session):
    """A TCPIP Session built on socket standard library using HiSLIP protocol."""

    # we don't decorate this class with Session.register() because we don't
    # want it to be registered in the _session_classes array, but we still
    # need to define session_type to make the set_attribute machinery work.
    session_type = (constants.InterfaceType.tcpip, "INSTR")

    # Override parsed to take into account the fact that this class is only used
    # for a specific kind of resource
    parsed: rname.TCPIPInstr

    @staticmethod
    def list_resources(wait_time=1.0) -> List[str]:
        resources = []
        try:
            for host in get_services("_hislip._tcp.local.", wait_time=wait_time):
                resources.append(f"TCPIP::{host}::hislip0,4880::INSTR")
        except NotImplementedError:
            warnings.warn(
                "TCPIP::hislip resource discovery requires the zeroconf package "
                "to be installed... try 'pip install zeroconf'",
                UserWarning,
            )
        return sorted(resources)

    def after_parsing(self) -> None:
        # TODO: board_number not handled

        if "," in self.parsed.lan_device_name:
            sub_address, port_str = self.parsed.lan_device_name.split(",")
            port = int(port_str)
        else:
            sub_address = self.parsed.lan_device_name
            port = 4880
        self.interface = hislip.Instrument(
            self.parsed.host_address,
            timeout=self.timeout,
            port=port,
            sub_address=sub_address,
        )

        # initialize the constant attributes
        self.attrs[ResourceAttribute.dma_allow_enabled] = constants.VI_FALSE
        self.attrs[ResourceAttribute.file_append_enabled] = constants.VI_FALSE
        self.attrs[ResourceAttribute.interface_instrument_name] = "TCPIP0 (HiSLIP)"
        self.attrs[ResourceAttribute.interface_number] = 0
        self.attrs[ResourceAttribute.io_prot] = constants.VI_PROT_NORMAL
        self.attrs[ResourceAttribute.read_buffer_operation_mode] = (
            constants.VI_FLUSH_DISABLE
        )
        self.attrs[ResourceAttribute.resource_lock_state] = constants.VI_NO_LOCK
        self.attrs[ResourceAttribute.send_end_enabled] = constants.VI_TRUE
        self.attrs[ResourceAttribute.suppress_end_enabled] = constants.VI_FALSE
        self.attrs[ResourceAttribute.tcpip_address] = self.parsed.host_address
        self.attrs[ResourceAttribute.tcpip_device_name] = self.parsed.lan_device_name
        self.attrs[ResourceAttribute.tcpip_hislip_overlap_enable] = constants.VI_FALSE
        self.attrs[ResourceAttribute.tcpip_hislip_version] = 0x0010_0000
        self.attrs[ResourceAttribute.tcpip_hostname] = self.parsed.host_address
        self.attrs[ResourceAttribute.tcpip_is_hislip] = constants.VI_TRUE
        self.attrs[ResourceAttribute.tcpip_nodelay] = constants.VI_TRUE
        self.attrs[ResourceAttribute.tcpip_port] = port
        self.attrs[ResourceAttribute.termchar] = ord("\n")
        self.attrs[ResourceAttribute.termchar_enabled] = constants.VI_FALSE
        self.attrs[ResourceAttribute.write_buffer_operation_mode] = (
            constants.VI_FLUSH_WHEN_FULL
        )

        # configure the variable attributes
        self.attrs[ResourceAttribute.tcpip_hislip_max_message_kb] = (
            self.get_max_message_kb,
            self.set_max_message_kb,
        )
        self.attrs[ResourceAttribute.tcpip_keepalive] = (
            self.get_keepalive,
            self.set_keepalive,
        )

        # TODO: additional attributes (someday)
        # self.attrs[ResourceAttribute.manufacturer_id] = 16711
        # self.attrs[ResourceAttribute.max_queue_length] = 50
        # self.attrs[ResourceAttribute.read_buffer_size] = 4096
        # self.attrs[ResourceAttribute.resource_impl_version] = 0x0050_0c01
        # self.attrs[ResourceAttribute.resource_manufacturer_id] = 4015
        # self.attrs[ResourceAttribute.resource_manufacturer_name] = 'Rohde & Schwarz GmbH'
        # self.attrs[ResourceAttribute.resource_spec_version] = 0x0050_0800
        # self.attrs[ResourceAttribute.user_data] = 0
        # self.attrs[ResourceAttribute.write_buffer_size] = 4096

    def get_max_message_kb(
        self, attribute: ResourceAttribute
    ) -> Tuple[int, StatusCode]:
        """Get the maximum HiSLIP message size in kilobytes."""
        max_msg_size_kb = int(round(self.interface.max_msg_size / 1024))
        return max_msg_size_kb, StatusCode.success

    def set_max_message_kb(
        self, attribute: ResourceAttribute, size_kb: int
    ) -> StatusCode:
        """Set the maximum HiSLIP message size in kilobytes."""
        if size_kb < 1:
            raise ValueError("size must be >= 1 kilobyte")

        if size_kb > 0xFFFF_FFFF:
            raise ValueError("size exceeds the range in the VISA spec")

        self.interface.max_msg_size = int(round(size_kb * 1024))
        return StatusCode.success

    def get_keepalive(self, attribute: ResourceAttribute) -> Tuple[bool, StatusCode]:
        """Is TCP keepalive enabled for the resource."""
        return self.interface.keepalive, StatusCode.success

    def set_keepalive(
        self, attribute: ResourceAttribute, keepalive: bool
    ) -> StatusCode:
        """Turns TCP keepalive on/off for this connection."""
        self.interface.keepalive = keepalive
        return StatusCode.success

    def close(self) -> StatusCode:
        self.interface.close()
        self.interface = None
        return StatusCode.success

    def _set_timeout(self, attribute: ResourceAttribute, value: int) -> StatusCode:
        status = super()._set_timeout(attribute, value)
        if hasattr(self.interface, "timeout"):
            self.interface.timeout = 1e-3 * value

        return status

    def read(self, count: int) -> Tuple[bytes, StatusCode]:
        """Reads data from device or interface synchronously.

        Corresponds to viRead function of the VISA library.

         Parameters
        -----------
        count : int
            Number of bytes to be read.

        Returns
        -------
        bytes
            Data read from the device
        StatusCode
            Return value of the library call.

        """
        try:
            data = self.interface.receive(count)
            status = (
                StatusCode.success_termination_character_read
                if self.interface._rmt
                else StatusCode.success_max_count_read
                if len(data) >= count
                else StatusCode.success
            )

        except socket.timeout:
            data, status = b"", StatusCode.error_timeout

        return data, status

    def write(self, data: bytes) -> Tuple[int, StatusCode]:
        """Writes data to device or interface synchronously.

        Corresponds to viWrite function of the VISA library.

        Parameters
        ----------
        data : bytes
            Data to be written.

        Returns
        -------
        int
            Number of bytes actually transferred
        StatusCode
            Return value of the library call.

        """
        self.interface.send(data)

        return len(data), StatusCode.success

    def clear(self) -> StatusCode:
        """Clears a device.

        Corresponds to viClear function of the VISA library.

        """
        self.interface.device_clear()

        return StatusCode.success

    def _get_attribute(self, attribute: ResourceAttribute) -> Tuple[Any, StatusCode]:
        """Get the value for a given VISA attribute for this session.

        Use to implement custom logic for attributes.

        Parameters
        ----------
        attribute : ResourceAttribute
            Attribute for which the state query is made

        Returns
        -------
        Any
            State of the queried attribute for a specified resource
        StatusCode
            Return value of the library call.

        """
        raise UnknownAttribute(attribute)

    def _set_attribute(
        self, attribute: ResourceAttribute, attribute_state: Any
    ) -> StatusCode:
        """Sets the state of an attribute.

        Corresponds to viSetAttribute function of the VISA library.

        Parameters
        ----------
        attribute : constants.ResourceAttribute
            Attribute for which the state is to be modified. (Attributes.*)
        attribute_state : Any
            The state of the attribute to be set for the specified object.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        raise UnknownAttribute(attribute)


class Vxi11CoreClient(vxi11.CoreClient):
    """
    make a connection using vxi11 protocol, optionally allowing the port number
    to be specified.  although in general the port number must be obtained by
    querying the portmapper, in practice any given instrument typically always
    uses the same port number.  this allows you to open that port on a firewall
    or set up an ssh tunnel to that port.

    """

    def __init__(
        self, host: str, port: Optional[int], open_timeout: Optional[int] = 5000
    ) -> None:
        self.packer = vxi11.Vxi11Packer()
        self.unpacker = vxi11.Vxi11Unpacker(b"")
        prog, vers = vxi11.DEVICE_CORE_PROG, vxi11.DEVICE_CORE_VERS

        if port is None:
            rpc.TCPClient.__init__(self, host, prog, vers, open_timeout)
        else:
            # bypass the portmapper lookup and use the specified port instead
            rpc.RawTCPClient.__init__(self, host, prog, vers, port, open_timeout)


class TCPIPInstrVxi11(Session):
    """A TCPIP Session built on socket standard library using VXI-11 protocol."""

    # we don't decorate this class with Session.register() because we don't
    # want it to be registered in the _session_classes array, but we still
    # need to define session_type to make the set_attribute machinery work.
    session_type = (constants.InterfaceType.tcpip, "INSTR")

    #: Maximum size of a chunk of data in bytes.
    max_recv_size: int

    #: Time to wait before erroring with a timeout when trying to acquire a lock
    lock_timeout: int

    #: Unique ID of the client used to authenticate messages.
    client_id: int

    #: ID of the link used for VXI-11 communication
    link: int

    # Override parsed to take into account the fact that this class is only used
    # for a specific kind of resource
    parsed: rname.TCPIPInstr

    # Setting if keepalive has been activated
    keepalive: bool

    @staticmethod
    def list_resources() -> List[str]:
        broadcast_addr = []
        if psutil is not None:
            # Get broadcast address for each interface
            for interface, snics in psutil.net_if_addrs().items():
                for snic in snics:
                    if snic.family is socket.AF_INET:
                        addr = snic.address
                        mask = snic.netmask
                        network = ipaddress.IPv4Network(addr + "/" + mask, strict=False)
                        broadcast_addr.append(str(network.broadcast_address))
        else:
            # If psutil unavailable fallback to default interface
            broadcast_addr.append("255.255.255.255")
            warnings.warn(
                "TCPIP:instr resource discovery is limited to the default interface."
                "Install psutil: pip install psutil if you want to scan all interfaces.",
                UserWarning,
            )

        pmap_list = [rpc.BroadcastUDPPortMapperClient(ip) for ip in broadcast_addr]
        for pmap in list(pmap_list):
            pmap.set_timeout(0)
            try:
                pmap.send_port(
                    (vxi11.DEVICE_CORE_PROG, vxi11.DEVICE_CORE_VERS, rpc.IPPROTO_TCP, 0)
                )
            except rpc.RPCError:
                pmap_list.remove(pmap)

        # Timeout for responses
        time.sleep(1)

        all_res = []
        for pmap in pmap_list:
            try:
                resp = pmap.recv_port(
                    (vxi11.DEVICE_CORE_PROG, vxi11.DEVICE_CORE_VERS, rpc.IPPROTO_TCP, 0)
                )
            except rpc.RPCError:
                pass
            else:
                res = [r[1][0] for r in resp if r[0] > 0]
                res = sorted(
                    res, key=lambda ip: tuple(int(part) for part in ip.split("."))
                )
                # TODO: Detect GPIB over TCPIP
                res = ["TCPIP::{}::INSTR".format(host) for host in res]
                all_res.extend(res)

        return all_res

    def after_parsing(self) -> None:
        # TODO: board_number not handled

        host_address = self.parsed.host_address
        if "," in host_address:
            host_address, port_str = host_address.split(",")
            port = int(port_str)
        else:
            port = None
        try:
            self.interface = Vxi11CoreClient(host_address, port, self.open_timeout)
        except rpc.RPCError:
            raise errors.VisaIOError(constants.VI_ERROR_RSRC_NFOUND)

        # vxi11 expect all timeouts to be expressed in ms and should be integers
        self.lock_timeout = 10000
        self.client_id = random.getrandbits(31)
        self.keepalive = False

        error, link, abort_port, max_recv_size = self.interface.create_link(
            self.client_id, 0, self.lock_timeout, self.parsed.lan_device_name
        )

        if error:
            raise Exception("error creating link: %d" % error)

        self.link = link
        self.max_recv_size = min(max_recv_size, 2**30)  # 1GB

        self.attrs[ResourceAttribute.tcpip_is_hislip] = False
        self.attrs[ResourceAttribute.tcpip_address] = self.parsed.host_address
        self.attrs[ResourceAttribute.tcpip_hostname] = ""
        self.attrs[ResourceAttribute.tcpip_device_name] = self.parsed.lan_device_name
        for name in ("SEND_END_EN", "TERMCHAR", "TERMCHAR_EN"):
            attribute = getattr(constants, "VI_ATTR_" + name)
            self.attrs[attribute] = attributes.AttributesByID[attribute].default

    def close(self) -> StatusCode:
        try:
            self.interface.destroy_link(self.link)
        except (errors.VisaIOError, socket.error, rpc.RPCError) as e:
            print("Error closing VISA link: {}".format(e))

        self.interface.close()
        self.link = 0
        self.interface = None

        return StatusCode.success

    def read(self, count: int) -> Tuple[bytes, StatusCode]:
        """Reads data from device or interface synchronously.

        Corresponds to viRead function of the VISA library.

        Parameters
        ----------
        count : int
            Number of bytes to be read.

        Returns
        -------
        bytes
            Data read
        StatusCode
            Return value of the library call.

        """
        if count < self.max_recv_size:
            chunk_length = count
        else:
            chunk_length = self.max_recv_size

        if self.get_attribute(ResourceAttribute.termchar_enabled)[0]:
            term_char, _ = self.get_attribute(ResourceAttribute.termchar)
            flags = vxi11.OP_FLAG_TERMCHAR_SET
        else:
            term_char = flags = 0

        read_data = bytearray()
        reason = 0
        # Stop on end of message or when a termination character has been
        # encountered.
        end_reason = vxi11.RX_END | vxi11.RX_CHR
        read_fun = self.interface.device_read
        status = StatusCode.success

        timeout = self._io_timeout
        start_time = time.time()
        while reason & end_reason == 0:
            # Decrease timeout so that the total timeout does not get larger
            # than the specified timeout.
            timeout = max(0, timeout - int((time.time() - start_time) * 1000))
            error, reason, data = read_fun(
                self.link, chunk_length, timeout, self.lock_timeout, flags, term_char
            )

            if error == vxi11.ErrorCodes.io_timeout:
                return bytes(read_data), StatusCode.error_timeout
            elif error:
                return bytes(read_data), StatusCode.error_io

            read_data.extend(data)
            count -= len(data)

            if count <= 0:
                status = StatusCode.success_max_count_read
                break

            chunk_length = min(count, chunk_length)

        return bytes(read_data), status

    def write(self, data: bytes) -> Tuple[int, StatusCode]:
        """Writes data to device or interface synchronously.

        Corresponds to viWrite function of the VISA library.

        Parameters
        ----------
        data : bytes
            Data to be written.

        Returns
        -------
        int
            Number of bytes actually transferred
        StatusCode
            Return value of the library call.

        """
        try:
            flags = 0
            num = len(data)
            offset = 0

            while num > 0:
                if num <= self.max_recv_size:
                    flags |= vxi11.OP_FLAG_END

                block = data[offset : offset + self.max_recv_size]

                error, size = self.interface.device_write(
                    self.link, self._io_timeout, self.lock_timeout, flags, block
                )

                if error == vxi11.ErrorCodes.io_timeout:
                    return offset, StatusCode.error_timeout

                elif error or size < len(block):
                    return offset, StatusCode.error_io

                offset += size
                num -= size

            return offset, StatusCode.success

        except vxi11.Vxi11Error:
            return 0, StatusCode.error_timeout

    def _get_attribute(self, attribute: ResourceAttribute) -> Tuple[Any, StatusCode]:
        """Get the value for a given VISA attribute for this session.

        Use to implement custom logic for attributes.

        Parameters
        ----------
        attribute :
            Resource attribute for which the state query is made

        Returns
        -------
        Any
            The state of the queried attribute for a specified resource
        StatusCode
            Return value of the library call.

        """
        # This is an abuse of the VISA standard
        if attribute == constants.VI_ATTR_TCPIP_KEEPALIVE:
            return self.keepalive, StatusCode.success

        elif attribute == constants.VI_ATTR_SUPPRESS_END_EN:
            raise NotImplementedError

        raise UnknownAttribute(attribute)

    def _set_attribute(
        self, attribute: ResourceAttribute, attribute_state: Any
    ) -> StatusCode:
        """Sets the state of an attribute.

        Corresponds to viSetAttribute function of the VISA library.

        Parameters
        ----------
        attribute : ResourceAttribute
             Attribute for which the state is to be modified.
        attribute_state : Any
            The state of the attribute to be set for the specified object.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """

        # In case of an environment with idle socket garbage collection (like docker)
        # sockets need to be kept alive. Set pyvisa.constants.ResourceAttribute.tcpip_keepalive to enable
        # keepalive packets even for VXI11 protocol. To read more on this issue
        # https://tech.xing.com/a-reason-for-unexplained-connection-timeouts-on-kubernetes-docker-abd041cf7e02
        if attribute == constants.VI_ATTR_TCPIP_KEEPALIVE:
            if attribute_state is True:
                self.interface.sock.setsockopt(
                    socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1
                )
                self.interface.sock.setsockopt(
                    socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 60
                )
                self.interface.sock.setsockopt(
                    socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 60
                )
                self.interface.sock.setsockopt(
                    socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 5
                )
                self.keepalive = True
            elif attribute_state is False:
                self.interface.sock.setsockopt(
                    socket.SOL_SOCKET, socket.SO_KEEPALIVE, 0
                )
                self.keepalive = False
            else:
                return StatusCode.error_nonsupported_format
            return StatusCode.success

        raise UnknownAttribute(attribute)

    def assert_trigger(self, protocol: constants.TriggerProtocol):
        """Asserts software or hardware trigger.

        Corresponds to viAssertTrigger function of the VISA library.

        Parameters
        ----------
        protocol : constants.TriggerProtocol
            Trigger protocol to use during assertion. Only default is supported.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        # XXX make this nicer (either validate protocol or pass it)
        error = self.interface.device_trigger(
            self.link, 0, self.lock_timeout, self._io_timeout
        )

        return VXI11_ERRORS_TO_VISA[error]

    def clear(self) -> StatusCode:
        """Clears a device.

        Corresponds to viClear function of the VISA library.

        """
        error = self.interface.device_clear(
            self.link, 0, self.lock_timeout, self._io_timeout
        )

        return VXI11_ERRORS_TO_VISA[error]

    def read_stb(self) -> Tuple[int, StatusCode]:
        """Reads a status byte of the service request.

        Corresponds to viReadSTB function of the VISA library.

        Returns
        -------
        int
            Service request status byte
        StatusCode
            Return value of the library call.

        """
        error, stb = self.interface.device_read_stb(
            self.link, 0, self.lock_timeout, self._io_timeout
        )

        return stb, VXI11_ERRORS_TO_VISA[error]

    def lock(
        self,
        lock_type: constants.Lock,
        timeout: int,
        requested_key: Optional[str] = None,
    ) -> Tuple[str, constants.StatusCode]:
        """Establishes an access mode to the specified resources.

        Corresponds to viLock function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        lock_type : constants.Lock
            Specifies the type of lock requested.
        timeout : int
            Absolute time period (in milliseconds) that a resource waits to get
            unlocked by the locking session before returning an error.
        requested_key : Optional[str], optional
            Requested locking key in the case of a shared lock. For an exclusive
            lock it should be None.

        Returns
        -------
        Optional[str]
            Key that can then be passed to other sessions to share the lock, or
            None for an exclusive lock.
        StatusCode
            Return value of the library call.

        """
        #  TODO: lock type not implemented
        flags = 0

        error = self.interface.device_lock(self.link, flags, self.lock_timeout)

        return "", VXI11_ERRORS_TO_VISA[error]

    def unlock(self) -> constants.StatusCode:
        """Relinquish a lock for the specified resource.

        Corresponds to viUnlock function of the VISA library.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        error = self.interface.device_unlock(self.link)

        return VXI11_ERRORS_TO_VISA[error]

    def _set_timeout(self, attribute: ResourceAttribute, value: int) -> StatusCode:
        """Sets timeout calculated value from python way to VI_ way"""
        if value == constants.VI_TMO_INFINITE:
            self.timeout = None
            self._io_timeout = 2**32 - 1
        elif value == constants.VI_TMO_IMMEDIATE:
            self.timeout = 0
            self._io_timeout = 0
        else:
            self.timeout = value / 1000.0
            self._io_timeout = int(self.timeout * 1000)
        return StatusCode.success


class TCPIPInstrVicp(Session):
    """VICP Session that uses pyvicp to do the low level communication."""

    # Override parsed to take into account the fact that this class is only used
    # for a specific kind of resource
    parsed: rname.VICPInstr

    @staticmethod
    def list_resources(wait_time=1.0) -> List[str]:
        resources = []
        try:
            services = get_services("_lxi._tcp.local.", wait_time=wait_time)
        except NotImplementedError:
            warnings.warn(
                "VICP resources discovery requires the zeroconf package to be "
                "installed... try 'pip install zeroconf'",
                UserWarning,
            )
            return []

        for host, properties in services.items():
            if properties["Manufacturer"].lower().startswith("lecroy"):
                resources.append(f"VICP::{host}::INSTR")

        return sorted(resources)

    def after_parsing(self) -> None:
        # TODO: board_number not handled
        if pyvicp is None:
            raise NotImplementedError(
                "VICP requires the pyvicp package to be installed... "
                "try 'pip install pyvicp'"
            )

        host_address = self.parsed.host_address
        if "," in host_address:
            host_address, port_str = host_address.split(",")
            port = int(port_str)
        else:
            port = 1861

        self.interface = pyvicp.Client(
            self.parsed.host_address, port, timeout=self.timeout
        )

        # initialize the constant attributes
        for name in ("SEND_END_EN", "TERMCHAR", "TERMCHAR_EN"):
            attribute = getattr(constants, "VI_ATTR_" + name)
            self.attrs[attribute] = attributes.AttributesByID[attribute].default

        self.attrs[ResourceAttribute.dma_allow_enabled] = constants.VI_FALSE
        self.attrs[ResourceAttribute.file_append_enabled] = constants.VI_FALSE
        self.attrs[ResourceAttribute.interface_instrument_name] = "TCPIP0 (VICP)"
        self.attrs[ResourceAttribute.interface_number] = 0
        self.attrs[ResourceAttribute.io_prot] = constants.VI_PROT_NORMAL
        self.attrs[ResourceAttribute.read_buffer_operation_mode] = (
            constants.VI_FLUSH_DISABLE
        )
        self.attrs[ResourceAttribute.resource_lock_state] = constants.VI_NO_LOCK
        self.attrs[ResourceAttribute.suppress_end_enabled] = constants.VI_FALSE
        self.attrs[ResourceAttribute.tcpip_address] = self.parsed.host_address
        self.attrs[ResourceAttribute.tcpip_hostname] = self.parsed.host_address
        self.attrs[ResourceAttribute.tcpip_is_hislip] = constants.VI_FALSE
        self.attrs[ResourceAttribute.tcpip_nodelay] = constants.VI_TRUE
        self.attrs[ResourceAttribute.tcpip_port] = port
        self.attrs[ResourceAttribute.write_buffer_operation_mode] = (
            constants.VI_FLUSH_WHEN_FULL
        )

        # configure the variable attributes
        self.attrs[ResourceAttribute.tcpip_keepalive] = (
            self.get_keepalive,
            self.set_keepalive,
        )

        # TODO: additional attributes (someday)
        # self.attrs[ResourceAttribute.manufacturer_id] = 16711
        # self.attrs[ResourceAttribute.max_queue_length] = 50
        # self.attrs[ResourceAttribute.read_buffer_size] = 4096
        # self.attrs[ResourceAttribute.resource_impl_version] = 0x0050_0c01
        # self.attrs[ResourceAttribute.resource_manufacturer_id] = 4015
        # self.attrs[ResourceAttribute.resource_manufacturer_name] = 'Rohde & Schwarz GmbH'
        # self.attrs[ResourceAttribute.resource_spec_version] = 0x0050_0800
        # self.attrs[ResourceAttribute.user_data] = 0
        # self.attrs[ResourceAttribute.write_buffer_size] = 4096

    def get_keepalive(self, attribute: ResourceAttribute) -> Tuple[bool, StatusCode]:
        """Is TCP keepalive enabled for the resource."""
        return self.interface.keepalive, StatusCode.success

    def set_keepalive(
        self, attribute: ResourceAttribute, keepalive: bool
    ) -> StatusCode:
        """Turns TCP keepalive on/off for this connection."""
        self.interface.keepalive = keepalive
        return StatusCode.success

    def close(self) -> StatusCode:
        self.interface.close()
        self.interface = None
        return StatusCode.success

    def read(self, count: int) -> Tuple[bytes, StatusCode]:
        """Reads data from device or interface synchronously.

        Corresponds to viRead function of the VISA library.

        Parameters
        -----------
        count : int
            Number of bytes to be read.

        Returns
        -------
        bytes
            Data read from the device
        StatusCode
            Return value of the library call.

        """
        try:
            data = self.interface.receive(count)
        except socket.timeout:
            return b"", StatusCode.error_timeout

        if len(data) >= count:
            return data, StatusCode.success_max_count_read
        else:
            return data, StatusCode.success_termination_character_read

    def write(self, data: bytes) -> Tuple[int, StatusCode]:
        """Writes data to device or interface synchronously.

        Corresponds to viWrite function of the VISA library.

        Parameters
        ----------
        data : bytes
            Data to be written.

        Returns
        -------
        int
            Number of bytes actually transferred
        StatusCode
            Return value of the library call.

        """
        self.interface.send(data)

        return len(data), StatusCode.success

    def clear(self) -> StatusCode:
        """Clears a device.

        Corresponds to viClear function of the VISA library.

        """
        self.interface.device_clear()

        return StatusCode.success

    def _set_timeout(self, attribute: ResourceAttribute, value: int) -> StatusCode:
        status = super()._set_timeout(attribute, value)
        if hasattr(self.interface, "timeout"):
            self.interface.timeout = 1e-3 * value

        return status

    def _get_attribute(self, attribute: ResourceAttribute) -> Tuple[Any, StatusCode]:
        """Get the value for a given VISA attribute for this session.

        Use to implement custom logic for attributes.

        Parameters
        ----------
        attribute : ResourceAttribute
            Attribute for which the state query is made

        Returns
        -------
        Any
            State of the queried attribute for a specified resource
        StatusCode
            Return value of the library call.

        """
        raise UnknownAttribute(attribute)

    def _set_attribute(
        self, attribute: ResourceAttribute, attribute_state: Any
    ) -> StatusCode:
        """Sets the state of an attribute.

        Corresponds to viSetAttribute function of the VISA library.

        Parameters
        ----------
        attribute : constants.ResourceAttribute
            Attribute for which the state is to be modified. (Attributes.*)
        attribute_state : Any
            The state of the attribute to be set for the specified object.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        raise UnknownAttribute(attribute)


if pyvicp is not None:
    Session.register(constants.InterfaceType.vicp, "INSTR")(TCPIPInstrVicp)
else:
    Session.register_unavailable(
        constants.InterfaceType.vicp,
        "INSTR",
        "Please install PyVICP to use this resource type.",
    )


@Session.register(constants.InterfaceType.tcpip, "SOCKET")
class TCPIPSocketSession(Session):
    """A TCPIP Session that uses the network standard library to do the low
    level communication.

    """

    # Details about implementation:
    # On Windows, select is not interrupted by KeyboardInterrupt, to avoid
    # blocking for very long time, we use a decreasing timeout in select.
    # A minimum select timeout which prevents using too short select interval
    # is also calculated and select timeout is not lower that that minimum
    # timeout. The absolute minimum is 1 ms as a consequence.
    # This is valid for connect and read operations

    #: Maximum size of a chunk of data in bytes.
    max_recv_size: int

    # Override parsed to take into account the fact that this class is only used
    # for a specific kind of resource
    parsed: rname.TCPIPSocket

    @staticmethod
    def list_resources() -> List[str]:
        # TODO: is there a way to get this?
        return []

    def after_parsing(self) -> None:
        # TODO: board_number not handled

        ret_status = self._connect()
        if ret_status != StatusCode.success:
            self.close()
            raise Exception("could not connect: {0}".format(str(ret_status)))

        self.max_recv_size = 4096
        # This buffer is used to store the bytes that appeared after
        # termination char
        self._pending_buffer = bytearray()

        self.attrs[ResourceAttribute.tcpip_address] = self.parsed.host_address
        self.attrs[ResourceAttribute.tcpip_port] = self.parsed.port
        self.attrs[ResourceAttribute.interface_number] = self.parsed.board
        self.attrs[ResourceAttribute.tcpip_nodelay] = (
            self._get_tcpip_nodelay,
            self._set_attribute,
        )
        self.attrs[ResourceAttribute.tcpip_hostname] = ""
        self.attrs[ResourceAttribute.tcpip_keepalive] = (
            self._get_tcpip_keepalive,
            self._set_tcpip_keepalive,
        )
        # to use default as ni visa driver (NI-VISA 15.0)
        self.attrs[ResourceAttribute.suppress_end_enabled] = True

        for name in ("TERMCHAR", "TERMCHAR_EN"):
            attribute = getattr(constants, "VI_ATTR_" + name)
            self.attrs[attribute] = attributes.AttributesByID[attribute].default

    def _connect(self) -> StatusCode:
        timeout = self.open_timeout / 1000.0 if self.open_timeout else 10.0
        try:
            self.interface = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.interface.setblocking(False)
            self.interface.connect_ex((self.parsed.host_address, int(self.parsed.port)))
        except Exception as e:
            raise Exception("could not connect: {0}".format(str(e)))
        finally:
            self.interface.setblocking(True)

        # minimum is in interval 100 - 500ms based on timeout
        min_select_timeout = max(min(timeout / 10.0, 0.5), 0.1)
        # initial 'select_timout' is half of timeout or max 2 secs
        # (max blocking time). min is from 'min_select_timeout'
        select_timout = max(min(timeout / 2.0, 2.0), min_select_timeout)
        # time, when loop shall finish
        finish_time = time.time() + timeout
        while True:
            # use select to wait for socket ready, max `select_timout` seconds
            r, w, x = select.select(
                [self.interface], [self.interface], [], select_timout
            )
            if self.interface in r or self.interface in w:
                return StatusCode.success

            if time.time() >= finish_time:
                # reached timeout
                return StatusCode.error_timeout

            # `select_timout` decreased to 50% of previous or
            # min_select_timeout
            select_timout = max(select_timout / 2.0, min_select_timeout)

    def close(self) -> StatusCode:
        self.interface.close()
        self.interface = None
        return StatusCode.success

    def read(self, count: int) -> Tuple[bytes, StatusCode]:
        """Reads data from device or interface synchronously.

        Corresponds to viRead function of the VISA library.

         Parameters
        -----------
        count : int
            Number of bytes to be read.

        Returns
        -------
        bytes
            Data read from the device
        StatusCode
            Return value of the library call.

        """
        if count < self.max_recv_size:
            chunk_length = count
        else:
            chunk_length = self.max_recv_size

        term_char, _ = self.get_attribute(ResourceAttribute.termchar)
        term_byte = common.int_to_byte(term_char) if term_char is not None else b""
        term_char_en, _ = self.get_attribute(ResourceAttribute.termchar_enabled)
        suppress_end_en, _ = self.get_attribute(ResourceAttribute.suppress_end_enabled)

        read_fun = self.interface.recv

        # minimum is in interval 1 - 100ms based on timeout, 1sec if no timeout
        # defined
        min_select_timeout = (
            1 if self.timeout is None else max(min(self.timeout / 100.0, 0.1), 0.001)
        )
        # initial 'select_timout' is half of timeout or max 2 secs
        # (max blocking time). min is from 'min_select_timeout'
        select_timout = (
            2.0
            if self.timeout is None
            else max(min(self.timeout / 2.0, 2.0), min_select_timeout)
        )
        # time, when loop shall finish, None means never ending story if no
        # data arrives
        finish_time = None if self.timeout is None else (time.time() + self.timeout)
        while True:
            # check, if we have any data received (from pending buffer or
            # further reading)
            if term_char_en and term_byte in self._pending_buffer:
                term_byte_index = self._pending_buffer.index(term_byte) + 1
                if term_byte_index > count:
                    term_byte_index = count
                    status = StatusCode.success_max_count_read
                else:
                    status = StatusCode.success_termination_character_read
                out = bytes(self._pending_buffer[:term_byte_index])
                self._pending_buffer = self._pending_buffer[term_byte_index:]
                return out, status

            if len(self._pending_buffer) >= count:
                out = bytes(self._pending_buffer[:count])
                self._pending_buffer = self._pending_buffer[count:]
                return out, StatusCode.success_max_count_read

            # use select to wait for read ready, max `select_timout` seconds
            r, w, x = select.select([self.interface], [], [], select_timout)

            read_data = b""
            if self.interface in r:
                read_data = read_fun(chunk_length)
                self._pending_buffer.extend(read_data)

            if not read_data:
                # can't read chunk or timeout
                if self._pending_buffer and not suppress_end_en:
                    # we have some data without termchar but no further data
                    # expected
                    out = bytes(self._pending_buffer[:count])
                    self._pending_buffer = self._pending_buffer[count:]
                    return out, StatusCode.success

                if finish_time and time.time() >= finish_time:
                    # reached timeout
                    out = bytes(self._pending_buffer[:count])
                    self._pending_buffer = self._pending_buffer[count:]
                    return out, StatusCode.error_timeout

                # `select_timout` decreased to 50% of previous or
                # min_select_timeout
                select_timout = max(select_timout / 2.0, min_select_timeout)

    def write(self, data: bytes) -> Tuple[int, StatusCode]:
        """Writes data to device or interface synchronously.

        Corresponds to viWrite function of the VISA library.

        Parameters
        ----------
        data : bytes
            Data to be written.

        Returns
        -------
        int
            Number of bytes actually transferred
        StatusCode
            Return value of the library call.

        """
        chunk_size = 4096

        num = sz = len(data)

        offset = 0

        while num > 0:
            block = data[offset : min(offset + chunk_size, sz)]

            try:
                # use select to wait for write ready
                select.select([], [self.interface], [])
                size = self.interface.send(block)
            except socket.timeout:
                return offset, StatusCode.error_io

            if size < len(block):
                return offset, StatusCode.error_io

            offset += size
            num -= size

        return offset, StatusCode.success

    def clear(self) -> StatusCode:
        """Clears a device.

        Corresponds to viClear function of the VISA library.

        """
        self._pending_buffer.clear()
        while True:
            r, w, x = select.select([self.interface], [], [], 0.1)
            if not r:
                break
            r[0].recv(4096)

        return StatusCode.success

    def flush(self, mask: BufferOperation) -> StatusCode:
        """Flush the specified buffers.
        Corresponds to viFlush function of the VISA library.
        Parameters
        ----------
        mask : constants.BufferOperation
            Specifies the action to be taken with flushing the buffer.
            The values can be combined using the | operator. However multiple
            operations on a single buffer cannot be combined.
        Returns
        -------
        constants.StatusCode
            Return value of the library call.
        """
        if mask & BufferOperation.discard_read_buffer:
            self.clear()
        if (
            mask & BufferOperation.discard_read_buffer_no_io
            or mask & BufferOperation.discard_receive_buffer
            or mask & BufferOperation.discard_receive_buffer2
        ):
            self._pending_buffer.clear()
        if (
            mask & BufferOperation.flush_write_buffer
            or mask & BufferOperation.flush_transmit_buffer
            or mask & BufferOperation.discard_write_buffer
            or mask & BufferOperation.discard_transmit_buffer
        ):
            pass

        return StatusCode.success

    def _get_tcpip_nodelay(
        self, attribute: ResourceAttribute
    ) -> Tuple[constants.VisaBoolean, StatusCode]:
        if self.interface:
            value = self.interface.getsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY)
            return (
                constants.VisaBoolean.true
                if value == 1
                else constants.VisaBoolean.false,
                StatusCode.success,
            )
        return constants.VisaBoolean.false, StatusCode.error_nonsupported_attribute

    def _set_tcpip_nodelay(
        self, attribute: ResourceAttribute, attribute_state: bool
    ) -> StatusCode:
        if self.interface:
            self.interface.setsockopt(
                socket.IPPROTO_TCP, socket.TCP_NODELAY, 1 if attribute_state else 0
            )
            return StatusCode.success
        return StatusCode.error_nonsupported_attribute

    def _get_tcpip_keepalive(
        self, attribute: ResourceAttribute
    ) -> Tuple[constants.VisaBoolean, StatusCode]:
        if self.interface:
            value = self.interface.getsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE)
            return (
                constants.VisaBoolean.true
                if value == 1
                else constants.VisaBoolean.false,
                StatusCode.success,
            )
        return constants.VisaBoolean.false, StatusCode.error_nonsupported_attribute

    def _set_tcpip_keepalive(
        self, attribute: ResourceAttribute, attribute_state: bool
    ) -> StatusCode:
        if self.interface:
            self.interface.setsockopt(
                socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1 if attribute_state else 0
            )
            self.interface.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 60)
            self.interface.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 60)
            self.interface.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 5)
            return StatusCode.success
        return StatusCode.error_nonsupported_attribute

    def _get_attribute(self, attribute: ResourceAttribute) -> Tuple[Any, StatusCode]:
        """Get the value for a given VISA attribute for this session.

        Use to implement custom logic for attributes.

        Parameters
        ----------
        attribute : ResourceAttribute
            Attribute for which the state query is made

        Returns
        -------
        Any
            State of the queried attribute for a specified resource
        StatusCode
            Return value of the library call.

        """
        raise UnknownAttribute(attribute)

    def _set_attribute(
        self, attribute: ResourceAttribute, attribute_state: Any
    ) -> StatusCode:
        """Sets the state of an attribute.

        Corresponds to viSetAttribute function of the VISA library.

        Parameters
        ----------
        attribute : constants.ResourceAttribute
            Attribute for which the state is to be modified. (Attributes.*)
        attribute_state : Any
            The state of the attribute to be set for the specified object.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        raise UnknownAttribute(attribute)


def get_services(service_type: str, wait_time: float = 0.1) -> Dict[str, dict]:
    if zeroconf is None:
        raise NotImplementedError(
            "Service discovery requires the zeroconf package to be installed... "
            "try 'pip install zeroconf'"
        )

    class MyListener(zeroconf.ServiceListener):
        def __init__(self, *args, **kwargs):
            self.services = {}
            super().__init__(*args, **kwargs)

        def remove_service(self, zc: zeroconf.Zeroconf, type_: str, name: str) -> None:
            del self.services[name]

        def add_service(self, zc: zeroconf.Zeroconf, type_: str, name: str) -> None:
            info = zc.get_service_info(type_, name)
            if info is None:
                return
            properties = {}
            for key, val in info.properties.items():
                if key == b"txtvers":
                    continue
                properties[key.decode()] = val.decode()
            ipaddr = ipaddress.ip_address(info.addresses[0])
            self.services[str(ipaddr)] = properties

        update_service = add_service

    zero_conf = zeroconf.Zeroconf()
    listener = MyListener()
    browser = zeroconf.ServiceBrowser(zero_conf, service_type, listener, delay=0)
    time.sleep(wait_time)
    browser.cancel()
    zero_conf.close()
    return listener.services



"""
Python implementation of HiSLIP protocol.  Based on the HiSLIP spec:

http://www.ivifoundation.org/downloads/Class%20Specifications/IVI-6.1_HiSLIP-1.1-2011-02-24.pdf
"""

import socket
import struct
import time
from typing import Dict, Optional, Tuple

PORT = 4880

MESSAGETYPE_STR: Dict[int, str] = {
    0: "Initialize",
    1: "InitializeResponse",
    2: "FatalError",
    3: "Error",
    4: "AsyncLock",
    5: "AsyncLockResponse",
    6: "Data",
    7: "DataEnd",
    8: "DeviceClearComplete",
    9: "DeviceClearAcknowledge",
    10: "AsyncRemoteLocalControl",
    11: "AsyncRemoteLocalResponse",
    12: "Trigger",
    13: "Interrupted",
    14: "AsyncInterrupted",
    15: "AsyncMaxMsgSize",
    16: "AsyncMaxMsgSizeResponse",
    17: "AsyncInitialize",
    18: "AsyncInitializeResponse",
    19: "AsyncDeviceClear",
    20: "AsyncServiceRequest",
    21: "AsyncStatusQuery",
    22: "AsyncStatusResponse",
    23: "AsyncDeviceClearAcknowledge",
    24: "AsyncLockInfo",
    25: "AsyncLockInfoResponse",
    26: "GetDescriptors",
    27: "GetDescriptorsResponse",
    28: "StartTLS",
    29: "AsyncStartTLS",
    30: "AsyncStartTLSResponse",
    31: "EndTLS",
    32: "AsyncEndTLS",
    33: "AsyncEndTLSResponse",
    34: "GetSaslMechanismList",
    35: "GetSaslMechanismListResponse",
    36: "AuthenticationStart",
    37: "AuthenticationExchange",
    38: "AuthenticationResult",
    # reserved for future use         39-127 inclusive
    # VendorSpecific                  128-255 inclusive
}
MESSAGETYPE: Dict[str, int] = {value: key for (key, value) in MESSAGETYPE_STR.items()}

FATALERRORMESSAGE: Dict[int, str] = {
    0: "Unidentified error",
    1: "Poorly formed message header",
    2: "Attempt to use connection without both channels established",
    3: "Invalid Initialization sequence",
    4: "Server refused connection due to maximum number of clients exceeded",
    5: "Secure connection failed",
    # 6-127:   reserved for HiSLIP extensions
    # 128-255: device defined errors
}
FATALERRORCODE: Dict[str, int] = {
    value: key for (key, value) in FATALERRORMESSAGE.items()
}

ERRORMESSAGE: Dict[int, str] = {
    0: "Unidentified error",
    1: "Unrecognized Message Type",
    2: "Unrecognized control code",
    3: "Unrecognized Vendor Defined Message",
    4: "Message too large",
    5: "Authentication failed",
    # 6-127:   Reserved
    # 128-255: Device defined errors
}
ERRORCODE: Dict[str, int] = {value: key for (key, value) in ERRORMESSAGE.items()}

LOCKCONTROLCODE: Dict[str, int] = {
    "release": 0,
    "request": 1,
}

LOCKRESPONSE: Dict[int, str] = {
    0: "failure",
    1: "success",  # or "success exclusive"
    2: "success shared",
    3: "error",
}

REMOTELOCALCONTROLCODE: Dict[str, int] = {
    "disableRemote": 0,
    "enableRemote": 1,
    "disableAndGTL": 2,
    "enableAndGotoRemote": 3,
    "enableAndLockoutLocal": 4,
    "enableAndGTRLLO": 5,
    "justGTL": 6,
}

HEADER_FORMAT = "!2sBBIQ"
# !  = network order,
# 2s = prologue ('HS'),
# B  = message type (unsigned byte),
# B  = control code (unsigned byte),
# I  = message parameter (unsigned int),
# Q  = payload length (unsigned long long)
HEADER_SIZE = struct.calcsize(HEADER_FORMAT)

DEFAULT_MAX_MSG_SIZE = 1 << 20  # from VISA spec


#########################################################################################


def receive_flush(sock: socket.socket, recv_len: int) -> None:
    """
    receive exactly 'recv_len' bytes from 'sock'.
    no explicit timeout is specified, since it is assumed
    that a call to select indicated that data is available.
    received data is thrown away and nothing is returned
    """
    # limit the size of the recv_buffer to something moderate
    # in order to limit the impact on virtual memory
    recv_buffer = bytearray(min(1 << 20, recv_len))
    bytes_recvd = 0

    while bytes_recvd < recv_len:
        request_size = min(len(recv_buffer), recv_len - bytes_recvd)
        data_len = sock.recv_into(recv_buffer, request_size)
        bytes_recvd += data_len


def receive_exact(sock: socket.socket, recv_len: int) -> bytes:
    """
    receive exactly 'recv_len' bytes from 'sock'.
    no explicit timeout is specified, since it is assumed
    that a call to select indicated that data is available.
    returns a bytearray containing the received data.
    """
    recv_buffer = bytearray(recv_len)
    receive_exact_into(sock, recv_buffer)
    return recv_buffer


def receive_exact_into(sock: socket.socket, recv_buffer: bytes) -> None:
    """
    receive data from 'sock' to exactly fill 'recv_buffer'.
    no explicit timeout is specified, since it is assumed
    that a call to select indicated that data is available.
    """
    view = memoryview(recv_buffer)
    recv_len = len(recv_buffer)
    bytes_recvd = 0

    while bytes_recvd < recv_len:
        request_size = recv_len - bytes_recvd
        data_len = sock.recv_into(view, request_size)
        bytes_recvd += data_len
        view = view[data_len:]

    if bytes_recvd > recv_len:
        raise MemoryError("socket.recv_into scribbled past end of recv_buffer")


def send_msg(
    sock: socket.socket,
    msg_type: str,
    control_code: int,
    message_parameter: Optional[int],
    payload: bytes = b"",
) -> None:
    """Send a message on sock w/ payload."""
    msg = bytearray(
        struct.pack(
            HEADER_FORMAT,
            b"HS",
            MESSAGETYPE[msg_type],
            control_code,
            message_parameter or 0,
            len(payload),
        )
    )
    # txdecode(msg, payload)  # uncomment for debugging
    msg.extend(payload)
    sock.sendall(msg)


class RxHeader:
    """Generic base class for receiving messages.

    specific protocol responses subclass this class.
    """

    def __init__(
        self,
        sock: socket.socket,
        expected_message_type: Optional[str] = None,
    ) -> None:
        """receive and decode the HiSLIP message header"""
        self.header = receive_exact(sock, HEADER_SIZE)
        # rxdecode(self.header)  # uncomment for debugging
        (
            prologue,
            msg_type,
            self.control_code,
            self.message_parameter,
            self.payload_length,
        ) = struct.unpack(HEADER_FORMAT, self.header)

        if prologue != b"HS":
            # XXX we should send a 'Fatal Error' to the server, close the
            # sockets, then raise an exception
            raise RuntimeError("protocol synchronization error")

        if msg_type not in MESSAGETYPE_STR:
            # XXX we should send 'Unrecognized message type' to the
            #     server and discard this packet plus any payload.
            raise RuntimeError("unrecognized message type: %d" % msg_type)

        self.msg_type = MESSAGETYPE_STR[msg_type]

        if expected_message_type is not None and self.msg_type != expected_message_type:
            # XXX we should send an 'Error: Unidentified Error' to the server
            # and discard this packet plus any payload
            payload = (
                (": " + str(receive_exact(sock, self.payload_length)))
                if self.payload_length > 0
                else b""
            )
            raise RuntimeError(
                "expected message type '%s', received '%s%s'"
                % (expected_message_type, self.msg_type, payload)
            )

        if self.msg_type == "DataEnd" or self.msg_type == "Data":
            assert self.control_code == 0
            self.message_id = self.message_parameter


class InitializeResponse(RxHeader):
    def __init__(self, sock: socket.socket) -> None:
        super().__init__(sock, "InitializeResponse")
        assert self.payload_length == 0
        self.overlap = bool(self.control_code)
        self.version, self.session_id = struct.unpack("!4xHH8x", self.header)


class AsyncInitializeResponse(RxHeader):
    def __init__(self, sock: socket.socket) -> None:
        super().__init__(sock, "AsyncInitializeResponse")
        assert self.control_code == 0
        assert self.payload_length == 0
        self.vendor_id = struct.unpack("!4x4s8x", self.header)


class AsyncMaxMsgSizeResponse(RxHeader):
    def __init__(self, sock: socket.socket) -> None:
        super().__init__(sock, "AsyncMaxMsgSizeResponse")
        assert self.control_code == 0
        assert self.message_parameter == 0
        assert self.payload_length == 8
        payload = receive_exact(sock, self.payload_length)
        self.max_msg_size = struct.unpack("!Q", payload)[0]


class AsyncDeviceClearAcknowledge(RxHeader):
    def __init__(self, sock: socket.socket) -> None:
        super().__init__(sock, "AsyncDeviceClearAcknowledge")
        self.feature_bitmap = self.control_code
        assert self.message_parameter == 0
        assert self.payload_length == 0


class AsyncInterrupted(RxHeader):
    def __init__(self, sock: socket.socket) -> None:
        super().__init__(sock, "AsyncInterrupted")
        assert self.control_code == 0
        self.message_id = self.message_parameter
        assert self.payload_length == 0


class AsyncLockInfoResponse(RxHeader):
    def __init__(self, sock: socket.socket) -> None:
        super().__init__(sock, "AsyncLockInfoResponse")
        self.exclusive_lock = self.control_code  # 0: no lock, 1: lock granted
        self.clients_holding_locks = self.message_parameter
        assert self.payload_length == 0


class AsyncLockResponse(RxHeader):
    def __init__(self, sock: socket.socket) -> None:
        super().__init__(sock, "AsyncLockResponse")
        self.lock_response = LOCKRESPONSE[self.control_code]
        assert self.message_parameter == 0
        assert self.payload_length == 0


class AsyncRemoteLocalResponse(RxHeader):
    def __init__(self, sock: socket.socket) -> None:
        super().__init__(sock, "AsyncRemoteLocalResponse")
        assert self.control_code == 0
        assert self.message_parameter == 0
        assert self.payload_length == 0


class AsyncServiceRequest(RxHeader):
    def __init__(self, sock: socket.socket) -> None:
        super().__init__(sock, "AsyncServiceRequest")
        self.server_status = self.control_code
        assert self.message_parameter == 0
        assert self.payload_length == 0


class AsyncStatusResponse(RxHeader):
    def __init__(self, sock: socket.socket) -> None:
        super().__init__(sock, "AsyncStatusResponse")
        self.server_status = self.control_code
        assert self.message_parameter == 0
        assert self.payload_length == 0


class DeviceClearAcknowledge(RxHeader):
    def __init__(self, sock: socket.socket) -> None:
        super().__init__(sock, "DeviceClearAcknowledge")
        self.feature_bitmap = self.control_code
        assert self.message_parameter == 0
        assert self.payload_length == 0


class Interrupted(RxHeader):
    def __init__(self, sock: socket.socket) -> None:
        super().__init__(sock, "Interrupted")
        assert self.control_code == 0
        self.message_id = self.message_parameter
        assert self.payload_length == 0


class Error(RxHeader):
    def __init__(self, sock: socket.socket) -> None:
        super().__init__(sock, "Error")
        self.error_code = ERRORMESSAGE[self.control_code]
        assert self.message_parameter == 0
        self.error_message = receive_exact(sock, self.payload_length)


class FatalError(RxHeader):
    def __init__(self, sock: socket.socket) -> None:
        super().__init__(sock, "FatalError")
        self.error_code = FATALERRORMESSAGE[self.control_code]
        assert self.message_parameter == 0
        self.error_message = receive_exact(sock, self.payload_length)


class Instrument:
    """
    this is the principal export from this module.  it opens up a HiSLIP
    connection to the instrument at the specified IP address.
    """

    def __init__(
        self,
        ip_addr: str,
        timeout: Optional[float] = None,
        port: int = PORT,
        sub_address: str = "hislip0",
    ) -> None:
        # init transaction:
        #     C->S: Initialize
        #     S->C: InitializeResponse
        #     C->S: AsyncInitialize
        #     S->C: AsyncInitializeResponse

        timeout = timeout or 5.0

        # open the synchronous socket and send an initialize packet
        self._sync = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sync.connect((ip_addr, port))
        self._sync.settimeout(timeout)
        self._sync.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        init = self.initialize(sub_address=sub_address.encode("ascii"))
        if init.overlap != 0:
            print("**** prefer overlap = %d" % init.overlap)

        # open the asynchronous socket and send an initialize packet
        self._async = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._async.connect((ip_addr, port))
        self._async.settimeout(timeout)
        self._async.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self._async_init = self.async_initialize(session_id=init.session_id)

        # initialize variables
        self.max_msg_size = DEFAULT_MAX_MSG_SIZE
        self.keepalive = False
        self.timeout = timeout
        self._rmt = 0
        self._message_id = 0xFFFF_FF00
        self._last_message_id: Optional[int] = None
        self._msg_type: str = ""
        self._payload_remaining: int = 0

    # ================ #
    # MEMBER FUNCTIONS #
    # ================ #

    def close(self) -> None:
        self._sync.close()
        self._async.close()

    @property
    def timeout(self) -> float:
        """Timeout value in seconds for both the sync and async sockets"""
        return self._timeout

    @timeout.setter
    def timeout(self, val: float) -> None:
        """Timeout value in seconds for both the sync and async sockets"""
        self._timeout = val
        self._sync.settimeout(self._timeout)
        self._async.settimeout(self._timeout)

    @property
    def max_msg_size(self) -> int:
        """Maximum HiSLIP message size in bytes."""
        return self._max_msg_size

    @max_msg_size.setter
    def max_msg_size(self, size: int) -> None:
        self._max_msg_size = self.async_maximum_message_size(size)

    @property
    def last_message_id(self) -> Optional[int]:
        return self._last_message_id

    @last_message_id.setter
    def last_message_id(self, message_id: Optional[int]) -> None:
        """Re-set last message id and related attributes"""
        self._last_message_id = message_id
        self._rmt = 0
        self._payload_remaining = 0
        self._msg_type = ""

    @property
    def keepalive(self) -> bool:
        """Status of the TCP keepalive.

        Keepalive is on/off for both the sync and async sockets

        If a connection is dropped as a result of “keepalives”, the error code
        VI_ERROR_CONN_LOST is returned to current and subsequent I/O
        calls on the session.

        """
        return self._keepalive

    @keepalive.setter
    def keepalive(self, keepalive: bool) -> None:
        self._keepalive = bool(keepalive)
        self._sync.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, bool(keepalive))
        self._async.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, bool(keepalive))

    def send(self, data: bytes) -> int:
        """Send the data on the synchronous channel.

        More than one packet may be necessary in order
        to not exceed max_payload_size.
        """
        # print(f"send({data=})")  # uncomment for debugging
        data_view = memoryview(data)
        num_bytes_to_send = len(data)
        max_payload_size = self._max_msg_size - HEADER_SIZE

        # send the data in chunks of max_payload_size bytes at a time
        while num_bytes_to_send > 0:
            if num_bytes_to_send <= max_payload_size:
                assert len(data_view) == num_bytes_to_send
                self._send_data_end_packet(data_view)
                bytes_sent = num_bytes_to_send
            else:
                self._send_data_packet(data_view[:max_payload_size])
                bytes_sent = max_payload_size

            data_view = data_view[bytes_sent:]
            num_bytes_to_send -= bytes_sent

        return len(data)

    def receive(self, max_len: int = 4096) -> bytes:
        """Receive data on the synchronous channel.

        Terminate after max_len bytes or after receiving a DataEnd message
        """

        # print(f"receive({max_len=})")  # uncomment for debugging

        # receive data, terminating after len(recv_buffer) bytes or
        # after receiving a DataEnd message.
        #
        # note the use of receive_exact_into (which calls socket.recv_into),
        # avoiding unnecessary copies.
        #
        recv_buffer = bytearray(max_len)
        view = memoryview(recv_buffer)
        bytes_recvd = 0

        while bytes_recvd < max_len:
            if self._payload_remaining <= 0:
                if self._msg_type == "DataEnd":
                    # truncate to the actual number of bytes received
                    recv_buffer = recv_buffer[:bytes_recvd]
                    break
                self._msg_type, self._payload_remaining = self._next_data_header()

            request_size = min(self._payload_remaining, max_len - bytes_recvd)
            receive_exact_into(self._sync, view[:request_size])
            self._payload_remaining -= request_size
            bytes_recvd += request_size
            view = view[request_size:]

        if bytes_recvd > max_len:
            raise MemoryError("scribbled past end of recv_buffer")

        # if there is no data remaining, set the RMT flag
        if self._payload_remaining == 0 and self._msg_type == "DataEnd":
            #
            # From IEEE Std 488.2: Response Message Terminator.
            #
            # RMT is the new-line accompanied by END sent from the server
            # to the client at the end of a response. Note that with HiSLIP
            # this is implied by the DataEND message.
            #
            self._rmt = 1

        return recv_buffer

    def _next_data_header(self) -> Tuple[str, int]:
        """
        receive the next data header (either Data or DataEnd), check the
        message_id, and return the msg_type and payload_length.
        """
        while True:
            header = RxHeader(self._sync)

            if header.msg_type in ("Data", "DataEnd"):
                # When receiving Data messages if the MessageID is not 0xffff ffff,
                # then verify that the MessageID indicated in the Data message is
                # the MessageID that the client sent to the server with the most
                # recent Data, DataEND or Trigger message.
                #
                # If the MessageIDs do not match, the client shall clear any Data
                # responses already buffered and discard the offending Data message

                if (
                    header.message_parameter == 0xFFFF_FFFF
                    or header.message_parameter == self.last_message_id
                ):
                    break

            # we're out of sync.  flush this message and continue.
            receive_flush(self._sync, header.payload_length)

        return header.msg_type, header.payload_length

    def device_clear(self) -> None:
        feature = self.async_device_clear()
        # Abandon pending messages and wait for in-process synchronous messages
        # to complete.
        time.sleep(0.1)
        # Indicate to server that synchronous channel is cleared out.
        self.device_clear_complete(feature)
        # reset messageID and resume normal opreation
        self._message_id = 0xFFFF_FF00

    def initialize(
        self,
        version: tuple = (1, 0),
        vendor_id: bytes = b"xx",
        sub_address: bytes = b"hislip0",
    ) -> InitializeResponse:
        """
        perform an Initialize transaction.
        returns the InitializeResponse header.
        """
        major, minor = version
        header = struct.pack(
            "!2sBBBB2sQ",
            b"HS",
            MESSAGETYPE["Initialize"],
            0,
            major,
            minor,
            vendor_id,
            len(sub_address),
        )
        # txdecode(header, sub_address)  # uncomment for debugging
        self._sync.sendall(header + sub_address)
        return InitializeResponse(self._sync)

    def async_initialize(self, session_id: int) -> AsyncInitializeResponse:
        """
        perform an AsyncInitialize transaction.
        returns the AsyncInitializeResponse header.
        """
        send_msg(self._async, "AsyncInitialize", 0, session_id)
        return AsyncInitializeResponse(self._async)

    def async_maximum_message_size(self, size: int) -> int:
        """
        perform an AsyncMaxMsgSize transaction.
        returns the max_msg_size from the AsyncMaxMsgSizeResponse packet.
        """
        # maximum_message_size transaction:
        #     C->S: AsyncMaxMsgSize
        #     S->C: AsyncMaxMsgSizeResponse
        payload = struct.pack("!Q", size)
        send_msg(self._async, "AsyncMaxMsgSize", 0, 0, payload)
        response = AsyncMaxMsgSizeResponse(self._async)
        return response.max_msg_size

    def async_lock_info(self) -> int:
        """
        perform an AsyncLockInfo transaction.
        returns the exclusive_lock from the AsyncLockInfoResponse packet.
        """
        # async_lock_info transaction:
        #     C->S: AsyncLockInfo
        #     S->C: AsyncLockInfoResponse
        send_msg(self._async, "AsyncLockInfo", 0, 0)
        response = AsyncLockInfoResponse(self._async)
        return response.exclusive_lock

    def async_lock_request(self, timeout: float, lock_string: str = "") -> str:
        """
        perform an AsyncLock request transaction.
        returns the lock_response from the AsyncLockResponse packet.
        """
        # async_lock transaction:
        #     C->S: AsyncLock
        #     S->C: AsyncLockResponse
        ctrl_code = LOCKCONTROLCODE["request"]
        timeout_ms = int(1e3 * timeout)
        send_msg(self._async, "AsyncLock", ctrl_code, timeout_ms, lock_string.encode())
        response = AsyncLockResponse(self._async)
        return response.lock_response

    def async_lock_release(self) -> str:
        """
        perform an AsyncLock release transaction.
        returns the lock_response from the AsyncLockResponse packet.
        """
        # async_lock transaction:
        #     C->S: AsyncLock
        #     S->C: AsyncLockResponse
        ctrl_code = LOCKCONTROLCODE["release"]
        send_msg(self._async, "AsyncLock", ctrl_code, self.last_message_id)
        response = AsyncLockResponse(self._async)
        return response.lock_response

    def async_remote_local_control(self, remotelocalcontrol: str) -> None:
        """
        perform an AsyncRemoteLocalControl transaction.
        """
        # remote_local transaction:
        #     C->S: AsyncRemoteLocalControl
        #     S->C: AsyncRemoteLocalResponse
        ctrl_code = REMOTELOCALCONTROLCODE[remotelocalcontrol]
        send_msg(
            self._async, "AsyncRemoteLocalControl", ctrl_code, self.last_message_id
        )
        AsyncRemoteLocalResponse(self._async)

    def async_status_query(self) -> int:
        """
        perform an AsyncStatusQuery transaction.
        returns the server_status from the AsyncStatusResponse packet.
        """
        # async_status_query transaction:
        #     C->S: AsyncStatusQuery
        #     S->C: AsyncStatusResponse
        send_msg(self._async, "AsyncStatusQuery", self._rmt, self._message_id)
        self._rmt = 0
        response = AsyncStatusResponse(self._async)
        return response.server_status

    def async_device_clear(self) -> int:
        """
        perform an AsyncDeviceClear transaction.
        returns the feature_bitmap from the AsyncDeviceClearAcknowledge packet.
        """
        send_msg(self._async, "AsyncDeviceClear", 0, 0)
        response = AsyncDeviceClearAcknowledge(self._async)
        return response.feature_bitmap

    def device_clear_complete(self, feature_bitmap: int) -> int:
        """
        perform a DeviceClear transaction.
        returns the feature_bitmap from the DeviceClearAcknowledge packet.
        """
        send_msg(self._sync, "DeviceClearComplete", feature_bitmap, 0)
        response = DeviceClearAcknowledge(self._sync)
        return response.feature_bitmap

    def trigger(self) -> None:
        """send a Trigger packet on the sync channel"""
        send_msg(self._sync, "Trigger", self._rmt, self._message_id)
        self.last_message_id = self._message_id
        self._message_id = (self._message_id + 2) & 0xFFFF_FFFF

    def _send_data_packet(self, payload: bytes) -> None:
        """send a Data packet on the sync channel"""
        send_msg(self._sync, "Data", self._rmt, self._message_id, payload)
        self.last_message_id = self._message_id
        self._message_id = (self._message_id + 2) & 0xFFFF_FFFF

    def _send_data_end_packet(self, payload: bytes) -> None:
        """send a DataEnd packet on the sync channel"""
        send_msg(self._sync, "DataEnd", self._rmt, self._message_id, payload)
        self.last_message_id = self._message_id
        self._message_id = (self._message_id + 2) & 0xFFFF_FFFF

    def fatal_error(self, error: str, error_message: str = "") -> None:
        err_msg = (error_message or error).encode()
        send_msg(self._sync, "FatalError", FATALERRORCODE[error], 0, err_msg)

    def error(self, error: str, error_message: str = "") -> None:
        err_msg = (error_message or error).encode()
        send_msg(self._sync, "Error", ERRORCODE[error], 0, err_msg)


# the following two routines are only used for debugging.
# they are commented out because their f-strings use a feature
# that is a syntax error in Python versions < 3.7

# def rxdecode(header):
#     (
#         prologue,
#         msg_type,
#         control_code,
#         message_parameter,
#         payload_length,
#     ) = struct.unpack(HEADER_FORMAT, header)
#
#     msg_type = MESSAGETYPE_STR[msg_type]
#     print(
#         f"Rx: {prologue=}, "
#         f"{msg_type=}, "
#         f"{control_code=}, "
#         f"{message_parameter=}, "
#         f"{payload_length=}"
#     )


# def txdecode(header, payload=b""):
#     (
#         prologue,
#         msg_type,
#         control_code,
#         message_parameter,
#         payload_length,
#     ) = struct.unpack(HEADER_FORMAT, header)
#
#     msg_type = MESSAGETYPE_STR[msg_type]
#     print(
#         f"Tx: {prologue=}, "
#         f"{msg_type=}, "
#         f"{control_code=}, "
#         f"{message_parameter=}, "
#         f"{payload_length=}, "
#         f"{len(payload)=}, "
#         f"{bytes(payload[:20]).decode('iso-8859-1')!r}"
#     )



# -*- coding: utf-8 -*-
"""Test the TCPIP based resources."""

import socket

import pytest

from pyvisa.constants import ResourceAttribute
from pyvisa.testsuite.keysight_assisted_tests import copy_func, require_virtual_instr
from pyvisa.testsuite.keysight_assisted_tests.test_tcpip_resources import (
    TestTCPIPInstr as TCPIPInstrBaseTest,
    TestTCPIPSocket as TCPIPSocketBaseTest,
)


@require_virtual_instr
class TestTCPIPInstr(TCPIPInstrBaseTest):
    """Test pyvisa-py against a TCPIP INSTR resource."""

    #: Type of resource being tested in this test case.
    #: See RESOURCE_ADDRESSES in the __init__.py file of this package for
    #: acceptable values
    RESOURCE_TYPE = "TCPIP::INSTR"

    #: Minimal timeout value accepted by the resource. When setting the timeout
    #: to VI_TMO_IMMEDIATE, Visa (Keysight at least) may actually use a
    #: different value depending on the values supported by the resource.
    MINIMAL_TIMEOUT = 0  # XXX should we try to have this match VISA ?

    # XXX Skip test clear to see if it has some bad side effect
    test_clear = pytest.mark.skip(copy_func(TCPIPInstrBaseTest.test_clear))

    test_wrapping_handler = pytest.mark.xfail(
        copy_func(TCPIPInstrBaseTest.test_wrapping_handler)
    )

    test_managing_visa_handler = pytest.mark.xfail(
        copy_func(TCPIPInstrBaseTest.test_managing_visa_handler)
    )

    test_wait_on_event = pytest.mark.xfail(
        copy_func(TCPIPInstrBaseTest.test_wait_on_event)
    )

    test_wait_on_event_timeout = pytest.mark.xfail(
        copy_func(TCPIPInstrBaseTest.test_wait_on_event_timeout)
    )

    test_getting_unknown_buffer = pytest.mark.xfail(
        copy_func(TCPIPInstrBaseTest.test_getting_unknown_buffer)
    )

    test_manual_async_read = pytest.mark.xfail(
        copy_func(TCPIPInstrBaseTest.test_manual_async_read)
    )

    test_uninstall_all_handlers = pytest.mark.xfail(
        copy_func(TCPIPInstrBaseTest.test_uninstall_all_handlers)
    )

    test_handler_clean_up_on_resource_del = pytest.mark.xfail(
        copy_func(TCPIPInstrBaseTest.test_handler_clean_up_on_resource_del)
    )

    test_uninstalling_missing_visa_handler = pytest.mark.xfail(
        copy_func(TCPIPInstrBaseTest.test_uninstalling_missing_visa_handler)
    )

    test_handling_invalid_handler = pytest.mark.xfail(
        copy_func(TCPIPInstrBaseTest.test_handling_invalid_handler)
    )

    test_write_raw_read_bytes = pytest.mark.xfail(
        copy_func(TCPIPInstrBaseTest.test_write_raw_read_bytes)
    )

    test_io_prot_attr = pytest.mark.xfail(
        copy_func(TCPIPInstrBaseTest.test_io_prot_attr)
    )

    test_shared_locking = pytest.mark.xfail(
        copy_func(TCPIPInstrBaseTest.test_shared_locking)
    )

    test_timeout = pytest.mark.xfail(copy_func(TCPIPInstrBaseTest.test_timeout))

    test_attribute_handling = pytest.mark.xfail(
        copy_func(TCPIPInstrBaseTest.test_attribute_handling)
    )

    def test_keepalive_attribute_vxi11(self):
        assert self.instr.visalib.sessions[self.instr.session].keepalive is False
        self.instr.set_visa_attribute(ResourceAttribute.tcpip_keepalive, True)
        assert self.instr.visalib.sessions[self.instr.session].keepalive is True
        assert (
            self.instr.visalib.sessions[self.instr.session].interface.sock.getsockopt(
                socket.SOL_SOCKET, socket.SO_KEEPALIVE
            )
            == 1
        )

        self.instr.set_visa_attribute(ResourceAttribute.tcpip_keepalive, False)
        assert self.instr.visalib.sessions[self.instr.session].keepalive is False
        assert (
            self.instr.visalib.sessions[self.instr.session].interface.sock.getsockopt(
                socket.SOL_SOCKET, socket.SO_KEEPALIVE
            )
            == 0
        )


@require_virtual_instr
class TestTCPIPSocket(TCPIPSocketBaseTest):
    """Test pyvisa-py against a TCPIP SOCKET resource."""

    #: Type of resource being tested in this test case.
    #: See RESOURCE_ADDRESSES in the __init__.py file of this package for
    #: acceptable values
    RESOURCE_TYPE = "TCPIP::SOCKET"

    #: Minimal timeout value accepted by the resource. When setting the timeout
    #: to VI_TMO_IMMEDIATE, Visa (Keysight at least) may actually use a
    #: different value depending on the values supported by the resource.
    MINIMAL_TIMEOUT = 1

    test_timeout = pytest.mark.xfail(copy_func(TCPIPSocketBaseTest.test_timeout))

    test_attribute_handling = pytest.mark.xfail(
        copy_func(TCPIPSocketBaseTest.test_attribute_handling)
    )

    test_stb = pytest.mark.xfail(copy_func(TCPIPSocketBaseTest.test_stb))
